# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from zef import *
from zef.ops import run
from .ariadne_utils import gqlify

#TODO types_dict will have collisions if  blocks of different types have same name i.e type Zeyad, enum Zeyad
def parse_schema(g, schema):
    """ Goes through the schema line by line initializing  entities based on their type.
        Those types are then connected to the root node.
        Depending on entity different handling takes place.

        During this process, a type_name to zefref is added to a dict which is then
        used by the connect_et_in_graph function to bind multiple instances of the same ET to
        one single zefref.
    """
    lines = schema.split('\n')
    root = instantiate(ET(gqlify("schema")), g)
    types_dict = {}

    for i in range(len(lines)):
        gql_type = ""
        if lines[i].lstrip(' ')[:4] == "type":  # i.e: type User {
            gql_type = "type"

        elif lines[i].lstrip(' ')[:5] == "input":  # i.e:  input CreateCar {
            gql_type = "input"

        elif lines[i].lstrip(' ')[:9] == "interface":  # i.e:   interface Node {
            gql_type = "interface"

        elif lines[i].lstrip(' ')[:6] == "scalar":  # i.e: scalar Date
            gql_type = "scalar"

        elif lines[i].lstrip(' ')[:4] == "enum":  # enum Availability {
            gql_type = "enum"

        if gql_type:
            name = gqlify(lines[i].strip().split(' ')[1])
            entity = initialize_named_entity(g, name, gql_type)
            instantiate(root, RT(gqlify(gql_type)), entity, g)
            types_dict[name] = entity

            if gql_type == "scalar":
                continue
            elif gql_type == "enum":
                i = handle_enum_type(g, name, entity, lines, i + 1)
            else:
                i = instantiate_type_fields(g, entity, lines, i + 1)

    connect_et_in_graph(g, schema, types_dict)
    return root, types_dict


def handle_enum_type(g, enum_type, entity, lines, i):
    line = lines[i]
    while "}" not in line:
        enum_val = line.strip(' ')
        EN(enum_type, enum_val)
        ae = instantiate(AET.Enum(enum_type), g)
        (ae <= EN(enum_type, enum_val)) | g | run
        r = instantiate(entity, RT(gqlify("field")), ae, g)
        i = i + 1
        line = lines[i]

    return i


def initialize_named_entity(g, name, gql_type):
    et_name = instantiate(AET.String, g)
    (et_name <= name) | g | run
    entity = instantiate(ET(gqlify(gql_type)), g)
    instantiate(entity, RT("Name"), et_name, g)
    return entity


def handle_list_type(g, field_type, aet_types):
    splitted = field_type.replace("]", "").replace("[", "").split("=")
    field_type = splitted[0].strip(" ")
    list_et = instantiate(ET(gqlify("list")), g)
    if field_type in aet_types:
        if len(splitted) > 1:
            default_values = splitted[1].split(",")
            for v in default_values:
                v = v.strip(" ")
                aet = instantiate(getattr(AET, field_type), g)
                aet <= type(aet_types[field_type])(v)
                instantiate(list_et, RT("argument"), aet, g)
        else:
            aet = instantiate(getattr(AET, field_type), g)
            instantiate(list_et, RT("argument"), aet, g)
    else:
        et_arg = instantiate(ET(gqlify(field_type)), g)
        instantiate(list_et, RT("argument"), et_arg, g)
    return list_et


def initialize_field_return(g, field_type):
    aet_types = {'String': "str", 'Int': 1, 'Float': 1.5, 'Boolean': True}
    #  i.e [String] or [String] = ["id"]
    if "[" in field_type:
        return handle_list_type(g, field_type, aet_types)

    #  i.e  Int = 2017 or Int
    splitted = field_type.split("=")
    field_type = splitted[0].strip(" ")

    if field_type in aet_types:
        if field_type == "Boolean":
            del(aet_types[field_type])
            field_type = "Bool"
            aet_types[field_type] = True
        aet = instantiate(getattr(AET, field_type), g)
        if len(splitted) > 1:
            (aet <= type(aet_types[field_type])(splitted[1].strip(" "))) | g | run
        return aet

    et = initialize_named_entity(g, field_type, "type")
    return et


def create_aet_relation(g, entity, relation_name, relation_target, relation_type):
    # relation between type and its field return
    stripped_target = relation_target.strip(' ').replace("!", "")
    field_return = initialize_field_return(g, stripped_target)
    r = instantiate(entity, RT(gqlify(relation_type)), field_return, g)
    if "!" in relation_target:
        instantiate(r, RT(gqlify("nonNullable")), instantiate(AET.Int, g), g)
    # relation between field relation and its field name
    name = instantiate(AET.String, g)
    (name <= relation_name) | g | run
    instantiate(r, RT("Name"), name, g)
    return r


def instantiate_type_fields(g, entity, lines, i):
    """ Instantiate GQL types as Entities on zef graph
    """
    line = lines[i]
    while "}" not in line:
        if "(" not in line:
            # i.e: username: String! or model: Int = 2017
            relation_name = gqlify(line[:line.index(':')])
            relation_target = line[line.index(':') + 1:]
            if not is_type(relation_target)[0]:
                create_aet_relation(g, entity, relation_name, relation_target, "field")

        i += 1
        line = lines[i]
    return i


def is_type(field_return):
    target_type = field_return.strip(' ').replace("]", "").replace("[", "").replace("!", "").replace(" ", "").split("=")
    if target_type[0].strip(' ') in ["String", "Int", "Float", "Boolean"]:
        return False, ""
    return True, target_type[0].strip(' ')


def create_et_relation(g, type_dict, et_name, entity, relation_target, relation_name, connection_type):
    field_entity = type_dict[gqlify(et_name)]
    if "[" in relation_target:
        list_et = instantiate(ET(gqlify("list")), g)
        instantiate(list_et, RT("argument"), field_entity, g)
        field_entity = list_et

    r = instantiate(entity, RT(gqlify(connection_type)), field_entity, g)
    if "!" in relation_target:
        instantiate(r, RT(gqlify("nonNullable")), instantiate(AET.Int, g), g)
    name = instantiate(AET.String, g)
    (name <= relation_name) | g | run
    instantiate(r, RT("Name"), name, g)
    return r


def connect_et_in_graph(g, schema, type_dict):
    """
        This function connects any non-AET field within a gql type, input, interface
        to an already instantiated zefref of that non-AET field.
    """
    lines = schema.split('\n')
    for i in range(len(lines)):
        if lines[i].lstrip(' ')[:4] == "type" \
                or lines[i].lstrip(' ')[:5] == "input" \
                or lines[i].lstrip(' ')[:9] == "interface":

            line = lines[i].strip().split(' ')
            name = gqlify(line[1])
            entity = type_dict[name]

            if "implements" in line:
                implement_entity = type_dict[gqlify(line[3].strip("{ "))]
                instantiate(entity, RT(gqlify("implements")), implement_entity, g)

            i = i + 1
            line = lines[i]
            while "}" not in line:
                if "(" not in line:
                    relation_name = gqlify(line[:line.index(':')])
                    relation_target = line[line.index(':') + 1:]
                    target_is_et, et_name = is_type(relation_target)
                    if target_is_et:
                        create_et_relation(g, type_dict, et_name, entity, relation_target, relation_name, "field")
                    # TODO: figure out why did we add this at some point!
                    # else:
                    #     create_aet_relation(g, entity, relation_name, relation_target, "field")
                else:
                    relation_name = gqlify(line[:line.index('(')])
                    relation_target = line[line.index('):') + 2:].strip(' ')
                    target_is_et, et_name = is_type(relation_target)
                    if target_is_et:
                        r = create_et_relation(g, type_dict, et_name, entity, relation_target, relation_name,
                                                    "field")
                    else:
                        r = create_aet_relation(g, entity, relation_name, relation_target, "field")
                    args = line[line.index('(') + 1: line.index(')')].split(',')
                    for arg in args:
                        relation_name = gqlify(arg[:arg.index(':')])  # i.e: GQL_ID
                        relation_target = arg[arg.index(':') + 1:]  # i.e: String
                        target_is_et, et_name = is_type(relation_target)
                        if target_is_et:
                            create_et_relation(g, type_dict, et_name, r, relation_target, relation_name,
                                                    "queryParams")
                        else:
                            create_aet_relation(g, r, relation_name, relation_target, "queryParams")

                i += 1
                line = lines[i]



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .components import *
from .zef_rich import show
from .visualizations import to_table, to_card

__all__ = [
    "show",
    "Text",
    "Code",
    "Style",
    "Table",
    "Column",
    "Frame",
    "HStack",
    "VStack",
    "BulletList",
    "NumberedList",
    "Paragraph",
    "to_table",
    "to_card",
]


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ctypes import *

BT = {
    0: "_unspecified",
    1: "ROOT_NODE",
    2: "TX_EVENT_NODE",
    3: "RAE_INSTANCE_EDGE",
    4: "TO_DELEGATE_EDGE",
    5: "NEXT_TX_EDGE",
    6: "ENTITY_NODE",
    7: "ATOMIC_ENTITY_NODE",
    8: "ATOMIC_VALUE_NODE",
    9: "RELATION_EDGE",
    10: "DELEGATE_INSTANTIATION_EDGE",
    11: "DELEGATE_RETIREMENT_EDGE",
    12: "INSTANTIATION_EDGE",
    13: "TERMINATION_EDGE",
    14: "ATOMIC_VALUE_ASSIGNMENT_EDGE",
    15: "DEFERRED_EDGE_LIST_NODE",
    16: "ASSIGN_TAG_NAME_EDGE",
    17: "NEXT_TAG_NAME_ASSIGNMENT_EDGE",
    18: "FOREIGN_GRAPH_NODE",
    19: "ORIGIN_RAE_EDGE",
    20: "ORIGIN_GRAPH_EDGE",
    21: "FOREIGN_ENTITY_NODE",
    22: "FOREIGN_ATOMIC_ENTITY_NODE",
    23: "FOREIGN_RELATION_EDGE",
}

def has_source_target(bt):
    return False
def extract_source_target(data, ptr):
    return ptr,{}

def has_edges(bt):
    return False
def extract_edges(data, ptr):
    return ptr,{}

def has_data_buffer(bt):
    return False
def extract_data_buffer(data, ptr):
    return ptr,{}

c_blob_index = c_int
# Note: need BaseUID as an array of bytes, as apparently C uses that for the alignment.
# c_BaseUID = c_longlong
c_BaseUID = c_ubyte*8
c_Time = c_double
c_TimeSlice = c_int
c_token = c_uint

def convert_BaseUID(raw):
    return ''.join('{:02x}'.format(x) for x in raw)

def convert_Time(raw):
    return {
        "zef_type": "Time",
        "seconds_since_1970": raw
    }

def convert_TimeSlice(raw):
    return {
        "zef_type": "TimeSlice",
        "slice": raw
    }

def convert_EntityType(raw):
    return {
        "zef_type": "EntityType",
        "entity_type_indx": raw
    }

def convert_AttributeEntityType(raw):
    return {
        "zef_type": "AttributeEntityType",
        "value": raw
    }

def convert_RelationType(raw):
    return {
        "zef_type": "RelationType",
        "relation_type_indx": raw
    }

def parse_version_0_1_0(data: bytes):
    start_index = 42
    index_stride = 16
    
    ptr = 0

    blobs = {}

    while True:
        if (ptr // index_stride) * index_stride != ptr:
            ptr = (ptr // index_stride + 1) * index_stride
        if ptr >= len(data):
            break
        index = start_index + (ptr // index_stride)
        assert ptr == (index - start_index) * index_stride, "ptr is not aligned with blob_index"

        bt = BT[int(data[ptr])]
        obj = {"type": bt}

        if bt not in extract_func:
            raise Exception(f"Don't know how to handle blob type {BT[bt]}")
        ptr,details = extract_func[bt](data, ptr)
        obj.update(details)

        # Other versions
        if has_source_target(bt):
            ptr,source_target = extract_source_target(data, ptr)
            obj.update(source_target)
        if has_edges(bt):
            ptr,edge_info = extract_edges(data, ptr)
            obj.update(edge_info)
        if has_data_buffer(bt):
            ptr,data_buffer = extract_data_buffer(data, ptr)
            obj.update(data_buffer)

        blobs[index] = obj

    assert ptr == len(data)

    blobs = combine_deferred_edge_lists(blobs)

    return blobs



def combine_deferred_edge_lists(blobs):
    out = {}
    for index,item in blobs.items():
        if item["type"] == "DEFERRED_EDGE_LIST_NODE":
            continue

        if "subsequent_deferred_edge_list_index" not in item:
            out[index] = item
            continue

        edge_list = []

        deferred_item = item
        while True:
            edge_list += deferred_item["local_edges"]
            if deferred_item["subsequent_deferred_edge_list_index"] == 0:
                break
            
            deferred_item = blobs[deferred_item["subsequent_deferred_edge_list_index"]]
            assert deferred_item["type"] == "DEFERRED_EDGE_LIST_NODE"

        new_item = {**item}
        new_item.pop("local_edges")
        new_item.pop("subsequent_deferred_edge_list_index")
        new_item["edges"] = edge_list

        out[index] = new_item

    return out



def extract_manual_edges(data, ptr, raw):
    size = raw.local_edge_indexes_capacity
    local_edges = (c_blob_index*size).from_buffer_copy(data, ptr)
    local_edges = [x for x in local_edges if x != 0]
    ptr += sizeof(c_blob_index)*size

    edge_details = {
        "local_edges": local_edges,
        "subsequent_deferred_edge_list_index": raw.subsequent_deferred_edge_list_index,
    }
    return ptr,edge_details

def extract_manual_buffer(data, ptr, raw):
    size = raw.buffer_size
    buffer = (c_ubyte*size).from_buffer_copy(data, ptr)
    ptr += size

    buffer_details = {
        "data_buffer": bytes(buffer),
    }
    return ptr,buffer_details

def extract_ROOT_NODE(data, ptr):
    class ROOT_NODE(LittleEndianStructure):
        _fields_ = [
            ("this_BlobType", c_ubyte),
            ("actual_written_data_layout_version_info_size", c_short),
            ("actual_written_graph_revision_info_size", c_short),
            ("uid", c_BaseUID),
            ("data_layout_version_info", c_byte*62),
            ("graph_revision_info", c_byte*64),
            ("local_edge_indexes_capacity", c_int),
            ("subsequent_deferred_edge_list_index", c_blob_index),
            ("edge_indexes", c_blob_index),
        ]

    raw = ROOT_NODE.from_buffer_copy(data, ptr)

    details = {
        "data_layout_version": raw.data_layout_version_info[:raw.actual_written_data_layout_version_info_size],
        "graph_revision": raw.graph_revision_info[:raw.actual_written_graph_revision_info_size],
        "uid": convert_BaseUID(raw.uid),
    }

    ptr += ROOT_NODE.edge_indexes.offset

    ptr,edge_details = extract_manual_edges(data, ptr, raw)
    details.update(edge_details)

    return ptr,details


def extract_TX_EVENT_NODE(data, ptr):
    class TX_EVENT_NODE(LittleEndianStructure):
        _fields_ = [
            ("this_BlobType", c_ubyte),
            ("local_edge_indexes_capacity", c_int),
            ("subsequent_deferred_edge_list_index", c_blob_index),
            ("time", c_Time),
            ("time_slice", c_TimeSlice),
            ("uid", c_BaseUID),
            ("edge_indexes", c_blob_index),
        ]

    raw = TX_EVENT_NODE.from_buffer_copy(data, ptr)

    details = {
        "uid": convert_BaseUID(raw.uid),
        "time": convert_Time(raw.time),
        "time_slice": convert_TimeSlice(raw.time_slice),
    }

    ptr += TX_EVENT_NODE.edge_indexes.offset

    ptr,edge_details = extract_manual_edges(data, ptr, raw)
    details.update(edge_details)

    return ptr,details


def extract_just_source_target(data, ptr):
    class NEXT_TX_EDGE(LittleEndianStructure):
        _fields_ = [
            ("this_BlobType", c_ubyte),
            ("source_node_index", c_blob_index),
            ("target_node_index", c_blob_index),
        ]

    raw = NEXT_TX_EDGE.from_buffer_copy(data, ptr)

    details = {
        "source_node_index": raw.source_node_index,
        "target_node_index": raw.target_node_index,
    }

    ptr += sizeof(NEXT_TX_EDGE)

    return ptr,details

def extract_RAE_INSTANCE_EDGE(data, ptr):
    class RAE_INSTANCE_EDGE(LittleEndianStructure):
        _fields_ = [
            ("this_BlobType", c_ubyte),
            ("local_edge_indexes_capacity", c_int),
            ("subsequent_deferred_edge_list_index", c_blob_index),
            ("source_node_index", c_blob_index),
            ("target_node_index", c_blob_index),
            ("edge_indexes", c_blob_index),
        ]

    raw = RAE_INSTANCE_EDGE.from_buffer_copy(data, ptr)

    details = {
        "source_node_index": raw.source_node_index,
        "target_node_index": raw.target_node_index,
    }

    ptr += RAE_INSTANCE_EDGE.edge_indexes.offset

    ptr,edge_details = extract_manual_edges(data, ptr, raw)
    details.update(edge_details)

    return ptr,details

def extract_TO_DELEGATE_EDGE(data, ptr):
    class TO_DELEGATE_EDGE(LittleEndianStructure):
        _fields_ = [
            ("this_BlobType", c_ubyte),
            ("local_edge_indexes_capacity", c_int),
            ("subsequent_deferred_edge_list_index", c_blob_index),
            ("source_node_index", c_blob_index),
            ("target_node_index", c_blob_index),
            ("edge_indexes", c_blob_index),
        ]

    raw = TO_DELEGATE_EDGE.from_buffer_copy(data, ptr)

    details = {
        "source_node_index": raw.source_node_index,
        "target_node_index": raw.target_node_index,
    }

    ptr += TO_DELEGATE_EDGE.edge_indexes.offset

    ptr,edge_details = extract_manual_edges(data, ptr, raw)
    details.update(edge_details)

    return ptr,details

def extract_ENTITY_NODE(data, ptr):
    class ENTITY_NODE(LittleEndianStructure):
        _fields_ = [
            ("this_BlobType", c_ubyte),
            ("local_edge_indexes_capacity", c_int),
            ("entity_type", c_token),
            ("subsequent_deferred_edge_list_index", c_blob_index),
            ("instantiation_time_slice", c_TimeSlice),
            ("termination_time_slice", c_TimeSlice),
            ("uid", c_BaseUID),
            ("edge_indexes", c_blob_index),
        ]

    raw = ENTITY_NODE.from_buffer_copy(data, ptr)

    details = {
        "entity_type": convert_EntityType(raw.entity_type),
        "instantiation_time_slice": convert_TimeSlice(raw.instantiation_time_slice),
        "termination_time_slice": convert_TimeSlice(raw.termination_time_slice),
        "uid": convert_BaseUID(raw.uid),
    }

    ptr += ENTITY_NODE.edge_indexes.offset

    ptr,edge_details = extract_manual_edges(data, ptr, raw)
    details.update(edge_details)

    return ptr,details


def extract_ATOMIC_ENTITY_NODE(data, ptr):
    class ATOMIC_ENTITY_NODE(LittleEndianStructure):
        _fields_ = [
            ("this_BlobType", c_ubyte),
            ("local_edge_indexes_capacity", c_int),
            ("my_atomic_entity_type", c_token),
            ("subsequent_deferred_edge_list_index", c_blob_index),
            ("instantiation_time_slice", c_TimeSlice),
            ("termination_time_slice", c_TimeSlice),
            ("uid", c_BaseUID),
            ("edge_indexes", c_blob_index),
        ]

    raw = ATOMIC_ENTITY_NODE.from_buffer_copy(data, ptr)

    details = {
        "my_atomic_entity_type": convert_AttributeEntityType(raw.my_atomic_entity_type),
        "instantiation_time_slice": convert_TimeSlice(raw.instantiation_time_slice),
        "termination_time_slice": convert_TimeSlice(raw.termination_time_slice),
        "uid": convert_BaseUID(raw.uid),
    }

    ptr += ATOMIC_ENTITY_NODE.edge_indexes.offset

    ptr,edge_details = extract_manual_edges(data, ptr, raw)
    details.update(edge_details)

    return ptr,details

def extract_ATOMIC_VALUE_NODE(data, ptr):
    class ATOMIC_VALUE_NODE(LittleEndianStructure):
        _fields_ = [
            ("this_BlobType", c_ubyte),
            ("my_atomic_entity_type", c_token),
            ("subsequent_deferred_edge_list_index", c_blob_index),
            ("buffer_size", c_uint),
            ("edge_indexes", c_blob_index),
            ("data_buffer", c_byte),
        ]

    raw = ATOMIC_VALUE_NODE.from_buffer_copy(data, ptr)
    local_edges = [x for x in raw.edge_indexes if x != 0]

    details = {
        "my_atomic_entity_type": convert_AttributeEntityType(raw.my_atomic_entity_type),
        "local_edges": local_edges,
    }

    ptr += ATOMIC_VALUE_NODE.data_buffer.offset

    ptr,buffer_details = extract_manual_buffer(data, ptr, raw)
    details.update(buffer_details)

    return ptr,details

def extract_RELATION_EDGE(data, ptr):
    class RELATION_EDGE(LittleEndianStructure):
        _fields_ = [
            ("this_BlobType", c_ubyte),
            ("hostage_flags", c_ubyte),
            ("local_edge_indexes_capacity", c_int),
            ("relation_type", c_token),
            ("subsequent_deferred_edge_list_index", c_blob_index),
            ("source_node_index", c_blob_index),
            ("target_node_index", c_blob_index),
            ("instantiation_time_slice", c_TimeSlice),
            ("termination_time_slice", c_TimeSlice),
            ("uid", c_BaseUID),
            ("edge_indexes", c_blob_index),
        ]

    raw = RELATION_EDGE.from_buffer_copy(data, ptr)

    details = {
        "hostage_flags": raw.hostage_flags,
        "relation_type": convert_RelationType(raw.relation_type),
        "instantiation_time_slice": convert_TimeSlice(raw.instantiation_time_slice),
        "termination_time_slice": convert_TimeSlice(raw.termination_time_slice),
        "source_node_index": raw.source_node_index,
        "target_node_index": raw.target_node_index,
        "uid": convert_BaseUID(raw.uid),
    }

    ptr += RELATION_EDGE.edge_indexes.offset

    ptr,edge_details = extract_manual_edges(data, ptr, raw)
    details.update(edge_details)

    return ptr,details

def extract_ATOMIC_VALUE_ASSIGNMENT_EDGE(data, ptr):
    class ATOMIC_VALUE_ASSIGNMENT_EDGE(LittleEndianStructure):
        _fields_ = [
            ("this_BlobType", c_ubyte),
            ("my_atomic_entity_type", c_token),
            ("buffer_size", c_uint),
            ("source_node_index", c_blob_index),
            ("target_node_index", c_blob_index),
            ("data_buffer", c_byte),
        ]

    raw = ATOMIC_VALUE_ASSIGNMENT_EDGE.from_buffer_copy(data, ptr)

    details = {
        "my_atomic_entity_type": convert_AttributeEntityType(raw.my_atomic_entity_type),
        "source_node_index": raw.source_node_index,
        "target_node_index": raw.target_node_index,
    }

    # ptr += ATOMIC_VALUE_ASSIGNMENT_EDGE.data_buffer.offset

    # There is a bug in the C code which means the size of the blob is bigger
    # than it actually is. Hence we need some custom logic here.
    buf_ptr = ptr + ATOMIC_VALUE_ASSIGNMENT_EDGE.data_buffer.offset

    _,buffer_details = extract_manual_buffer(data, buf_ptr, raw)
    details.update(buffer_details)

    ptr += sizeof(ATOMIC_VALUE_ASSIGNMENT_EDGE) - 1 + len(buffer_details["data_buffer"]) + 1

    return ptr,details

def extract_DEFERRED_EDGE_LIST_NODE(data, ptr):
    class DEFERRED_EDGE_LIST_NODE(LittleEndianStructure):
        _fields_ = [
            ("this_BlobType", c_ubyte),
            ("local_edge_indexes_capacity", c_int),
            ("preceding_edge_list_index", c_blob_index),
            ("subsequent_deferred_edge_list_index", c_blob_index),
            ("edge_indexes", c_blob_index),
        ]

    raw = DEFERRED_EDGE_LIST_NODE.from_buffer_copy(data, ptr)

    details = {}

    ptr += DEFERRED_EDGE_LIST_NODE.edge_indexes.offset

    ptr,edge_details = extract_manual_edges(data, ptr, raw)
    details.update(edge_details)

    return ptr,details

def extract_ASSIGN_TAG_NAME_EDGE(data, ptr):
    class ASSIGN_TAG_NAME_EDGE(LittleEndianStructure):
        _fields_ = [
            ("this_BlobType", c_ubyte),
            ("chars_used_in_buffer", c_short),
            ("local_edge_indexes_capacity", c_int),
            ("subsequent_deferred_edge_list_index", c_blob_index),
            ("source_node_index", c_blob_index),
            ("target_node_index", c_blob_index),
            ("rel_ent_tag_name_buffer", (c_byte*50)),
            ("edge_indexes", c_blob_index),
        ]

    raw = ASSIGN_TAG_NAME_EDGE.from_buffer_copy(data, ptr)

    details = {
        "source_node_index": raw.source_node_index,
        "target_node_index": raw.target_node_index,
        "data_buffer": bytes(raw.rel_ent_tag_name_buffer[:raw.chars_used_in_buffer]),
    }

    ptr += ASSIGN_TAG_NAME_EDGE.edge_indexes.offset

    ptr,edge_details = extract_manual_edges(data, ptr, raw)
    details.update(edge_details)

    return ptr,details


def extract_FOREIGN_GRAPH_NODE(data, ptr):
    class FOREIGN_GRAPH_NODE(LittleEndianStructure):
        _fields_ = [
            ("this_BlobType", c_ubyte),
            ("local_edge_indexes_capacity", c_int),
            ("internal_foreign_graph_index", c_int),
            ("subsequent_deferred_edge_list_index", c_blob_index),
            ("uid", c_BaseUID),
            ("edge_indexes", c_blob_index),
        ]

    raw = FOREIGN_GRAPH_NODE.from_buffer_copy(data, ptr)

    details = {
        "internal_foreign_graph_index": raw.internal_foreign_graph_index,
        "uid": convert_BaseUID(raw.uid),
    }

    ptr += FOREIGN_GRAPH_NODE.edge_indexes.offset

    ptr,edge_details = extract_manual_edges(data, ptr, raw)
    details.update(edge_details)

    return ptr,details

def extract_FOREIGN_ENTITY_NODE(data, ptr):
    class FOREIGN_ENTITY_NODE(LittleEndianStructure):
        _fields_ = [
            ("this_BlobType", c_ubyte),
            ("local_edge_indexes_capacity", c_int),
            ("entity_type", c_token),
            ("subsequent_deferred_edge_list_index", c_blob_index),
            ("uid", c_BaseUID),
            ("edge_indexes", c_blob_index),
        ]

    raw = FOREIGN_ENTITY_NODE.from_buffer_copy(data, ptr)

    details = {
        "entity_type": convert_EntityType(raw.entity_type),
        "uid": convert_BaseUID(raw.uid),
    }

    ptr += FOREIGN_ENTITY_NODE.edge_indexes.offset

    ptr,edge_details = extract_manual_edges(data, ptr, raw)
    details.update(edge_details)

    return ptr,details

def extract_FOREIGN_ATOMIC_ENTITY_NODE(data, ptr):
    class FOREIGN_ATOMIC_ENTITY_NODE(LittleEndianStructure):
        _fields_ = [
            ("this_BlobType", c_ubyte),
            ("local_edge_indexes_capacity", c_int),
            ("atomic_entity_type", c_token),
            ("subsequent_deferred_edge_list_index", c_blob_index),
            ("uid", c_BaseUID),
            ("edge_indexes", c_blob_index),
        ]

    raw = FOREIGN_ATOMIC_ENTITY_NODE.from_buffer_copy(data, ptr)

    details = {
        "entity_type": convert_EntityType(raw.entity_type),
        "uid": convert_BaseUID(raw.uid),
    }

    ptr += FOREIGN_ATOMIC_ENTITY_NODE.edge_indexes.offset

    ptr,edge_details = extract_manual_edges(data, ptr, raw)
    details.update(edge_details)

    return ptr,details

def extract_FOREIGN_RELATION_EDGE(data, ptr):
    class FOREIGN_RELATION_EDGE(LittleEndianStructure):
        _fields_ = [
            ("this_BlobType", c_ubyte),
            ("local_edge_indexes_capacity", c_int),
            ("relation_type", c_token),
            ("source_node_index", c_blob_index),
            ("target_node_index", c_blob_index),
            ("subsequent_deferred_edge_list_index", c_blob_index),
            ("uid", c_BaseUID),
            ("edge_indexes", c_blob_index),
        ]

    raw = FOREIGN_RELATION_EDGE.from_buffer_copy(data, ptr)

    details = {
        "relation_type": convert_RelationType(raw.relation_type),
        "uid": convert_BaseUID(raw.uid),
        "source_node_index": raw.source_node_index,
        "target_node_index": raw.target_node_index,
    }

    ptr += FOREIGN_RELATION_EDGE.edge_indexes.offset

    ptr,edge_details = extract_manual_edges(data, ptr, raw)
    details.update(edge_details)

    return ptr,details


extract_func = {
    "ROOT_NODE": extract_ROOT_NODE,
    "TX_EVENT_NODE": extract_TX_EVENT_NODE,
    "NEXT_TX_EDGE": extract_just_source_target,
    "RAE_INSTANCE_EDGE": extract_RAE_INSTANCE_EDGE,
    "TO_DELEGATE_EDGE": extract_TO_DELEGATE_EDGE,
    "ENTITY_NODE": extract_ENTITY_NODE,
    "ATOMIC_ENTITY_NODE": extract_ATOMIC_ENTITY_NODE,
    "ATOMIC_VALUE_NODE": extract_ATOMIC_VALUE_NODE,
    "RELATION_EDGE": extract_RELATION_EDGE,
    "DELEGATE_INSTANTIATION_EDGE": extract_just_source_target,
    "DELEGATE_RETIREMENT_EDGE": extract_just_source_target,
    "INSTANTIATION_EDGE": extract_just_source_target,
    "TERMINATION_EDGE": extract_just_source_target,
    "ATOMIC_VALUE_ASSIGNMENT_EDGE": extract_ATOMIC_VALUE_ASSIGNMENT_EDGE,
    "DEFERRED_EDGE_LIST_NODE": extract_DEFERRED_EDGE_LIST_NODE,
    "ASSIGN_TAG_NAME_EDGE": extract_ASSIGN_TAG_NAME_EDGE,
    "FOREIGN_GRAPH_NODE": extract_FOREIGN_GRAPH_NODE,
    "ORIGIN_RAE_EDGE": extract_just_source_target,
    "ORIGIN_GRAPH_EDGE": extract_just_source_target,
    "FOREIGN_ENTITY_NODE": extract_FOREIGN_ENTITY_NODE,
    "FOREIGN_ATOMIC_ENTITY_NODE": extract_FOREIGN_ATOMIC_ENTITY_NODE,
    "FOREIGN_RELATION_EDGE": extract_FOREIGN_RELATION_EDGE,
}







# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "subgraph",
]

from ..core import *
from ..ops import *


def subgraph(gs, ets=ET, rts=RT):
    """Produce a simple directed graph with nodes and edges that represents a
    subgraph of a given GraphSlice `gs`.
    
    Given a set of EntityTypes `ets` to interpret as nodes, and a set of
    RelationTypes `rts` to interpret as edges, builds a list `nodes` which maps
    each index of the list to a ZefRef of the entity, and an adjacency matrix
    `A` that maps edges to the ZefRef of the relation.
    
    The adjacency matrix returned is a sparse matrix containing integers where
    an edge is present, and those integers are themselves indices into an
    additional list, `edges`, which maps edges to their original ZefRef.

    Note: due to limitations of the scipy.sparse package, index 0 cannot be used
    as an index. As a workaround, we set `edges[0] = None`, and the
    first true edge is at `edges[1]`.

    Return signature is `(A, nodes, edges)`
    
    If `ets` is a single EntityType, it will be automatically promoted to a
    list. Similarly with `rts`.

    """
    from scipy.sparse import csc_matrix, lil_matrix

    # if isinstance(ets, EntityType):
    #     ets = [ets]
    # if isinstance(rts, RelationType):
    #     rts = [rts]

    next_edge_ind = 1

    # nodes = gs | all[ET] | filter[rae_type | contained_in[ets]] | func[list] | collect
    nodes = gs | all[ets] | func[list] | collect
    edges = [None]

    N = len(nodes)
    A = lil_matrix((N,N), dtype=int)
    for i,i_node in enumerate(nodes):
        for edge in (i_node > L[RT]
                     | filter[is_a[rts]]
                     | filter[target | contained_in[nodes]]):
            j = nodes.index(target(edge))

            A[i,j] = next_edge_ind
            next_edge_ind += 1
            print("In here", i, j, A)
            edges += [edge]

    # A = A.tocsc()
    return A, nodes, edges


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ... import *
from ...ops import *

import builtins

import networkx
from networkx.classes.digraph import DiGraph

def split_by_func(s: str, func):
    all_lists = []
    while s:
        for ind in range(len(s)):
            if func(s[ind]):
                break
        else:
            ind = ind+1
        all_lists.append(s[:ind])
        s = s[ind+1:]

    import itertools
    return list(builtins.filter(len, all_lists))
            
def default_ET_translation(name, data):
    # By default, split by anything that is punctuation or a number and turn that into words.
    words = split_by_func(name, lambda c: not c.isalpha())

    return ''.join(builtins.map(lambda w: w.capitalize(), words))

def default_RT_translation(conn, data):
    # Use the target as the RT name
    return default_ET_translation(conn[1], data)

def default_fieldname_translation(name):
    # Numbers can be important in fields, so don't lose them.
    words = split_by_func(name, lambda c: not c.isalnum())

    return ''.join(builtins.map(lambda w: w.capitalize(), words))

def default_aet_translation(name, val):
    if type(val) == bool:
        return AET.Bool
    if type(val) == int:
        return AET.Int
    if type(val) == float:
        return AET.Float
    if type(val) == str:
        return AET.String
    
    
    

def inject_networkx_into_zef(nxg: DiGraph,
                             zg: Graph,
                             ET_translation = default_ET_translation,
                             RT_translation = default_RT_translation,
                             fieldname_translation = default_fieldname_translation,
                             aet_translation = default_aet_translation,
                             ignore_excess_ET_RT_types: bool = False,
                             ignore_duplicate_fields: bool = False,
                             # ET to use when the translation returns an empty string
                             default_et = ET.Node,
                             # RT to use when the translation returns an empty string
                             default_rt = RT.Edge,
                             store_id = RT.ID,
                             ):
    # First check that we don't end up with a large number of ET and RT types.
    # This would be likely due to a bad conversion of the names
    if not ignore_excess_ET_RT_types:
        theset = set()
        for pair in nxg.nodes.items():
            theset.add(ET_translation(*pair))
        if len(nxg.nodes) > 12 and len(theset) > len(nxg.nodes) / 3:
            raise Exception(f"After converting the node names, there were still {len(theset)} types left out of {len(nxg.nodes)} nodes. If this is expected, then you can disable this check by passing `ignore_excess_ET_RT_types=True` as a keyword to `inject_networkx_into_zef`.")

        theset = set()
        for pair in nxg.edges.items():
            theset.add(RT_translation(*pair))
        if len(nxg.edges) > 12 and len(theset) > len(nxg.edges) / 3:
            raise Exception(f"After converting the edge names, there were still {len(theset)} types left out of {len(nxg.edges)} edges. If this is expected, then you can disable this check by passing `ignore_excess_ET_RT_types=True` as a keyword to `inject_networkx_into_zef`.")


    with Transaction(zg):
        entity_mapping = {}
        relation_mapping = {}
        for key,data in nxg.nodes.items():
            try:
                name = ET_translation(key,data)
                if name == "":
                    et = default_et
                else:
                    et = ET(name)
            except Exception as exc:
                raise Exception(f"Unable to make ET from '{name}'") from exc

            ent = et | zg | run
            entity_mapping[key] = ent

            realise_fields(zg, ent, data, fieldname_translation, aet_translation)

        for conn,data in nxg.edges.items():
            try:
                name = RT_translation(conn, data)
                if name == "":
                    rt = default_rt
                else:
                    rt = RT(name)
            except Exception as exc:
                raise Exception(f"Unable to make RT from '{name}'") from exc

            source = entity_mapping[conn[0]]
            target = entity_mapping[conn[1]]
            _,rel,_ = (source, rt, target) | zg | run

            relation_mapping[conn] = rel

            realise_fields(zg, rel, data, fieldname_translation, aet_translation)

    return entity_mapping, relation_mapping
       
        
        
def realise_fields(zg: Graph,
                   obj: ZefRef,
                   data: dict,
                   fieldname_translation = default_fieldname_translation,
                   aet_translation = default_aet_translation,
                   ignore_duplicate_fields: bool = False):
    
    # A check - don't want accidental duplicates in the field translation, unless specifically allowed.
    # I don't think this is even possible in networkx as the fields are represented with dictionaries.
    if not ignore_duplicate_fields:
        assert(len(set(builtins.map(fieldname_translation, data.keys()))) == len(data.keys()))

    for name,val in data.items():
        rt = RT(fieldname_translation(name))
        aet = aet_translation(name, val)
        if aet is None:
            raise Exception(f"Unknown type for field with name {name} and value {val}")
        z_val = (aet <= val) | zg | run
        (obj, rt, z_val) | zg | run



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ariadne import make_executable_schema, graphql_sync
from . import schema_graph_to_file as file_generator
from . import schema_file_to_graph as graph_generator
import os
from zef import Graph, RT
from zef.ops import *
import json


def generate_resolvers_fcts(schema_root, resolvers_destination):
    """ Generates ariadne resolvers from the generate_ariadne_resolvers.py
        cog file. And outputs generated_resolvers.py

        We pass a global dict with the root node of the schema graph that cog
        uses to generate the resolvers.
    """
    from cogapp import Cog
    cog = Cog()
    cog.options.bReplace = True
    optional = single_or[None]

    z = schema_root | Outs[RT.DefaultResolversList] | optional | maybe_value | collect
    if z is None:
        default_resolvers_list = []
    else:
        default_resolvers_list = from_json(z)
        
        
    z = schema_root | Outs[RT.FallbackResolvers] | optional | collect
    if z is None:
        fallback_resolvers = lambda *args,**kwds: ("return None #This means that no resolver is defined!", ["z", "ctx"])
    else:
        fallback_resolvers = z | value | collect
        d = {}
        exec(fallback_resolvers, d)
        fallback_resolvers = d["fallback_resolvers"]

    additional_exec = schema_root | Outs[RT.AdditionalExec] | optional | value_or[""] | collect

    global_dict = {"schema_root": schema_root,
                   "default_resolvers_list": default_resolvers_list,
                   "fallback_resolvers": fallback_resolvers,
                   "additional_exec": additional_exec}

    try:
        path = os.path.dirname(os.path.realpath(__file__))
        cog.processFile(os.path.join(path, "generate_ariadne_resolvers.py"),
                        os.path.join(resolvers_destination, "ariadne_resolvers.py"), globals=global_dict)
    except Exception as exc:
        print(f'An exception was raised when processing file {exc}')


@func
def generate_api_schema_file(schema_root, file_path):
    """
    Given a ZefRef to a schema_root, a file, defined in the file_path,
    containing a 'generated_schema' string variable is generated.

    If the requirement is to just produce a string holding the schema, this effect could be used:

    Effect({
        type: FX.GraphQL.GenerateSchemaString,
        schema_root: schema_root,
    })
    """
    file_generator.generate_api_schema_file(schema_root, file_path)


@func
def generate_graph_from_file(schema_str, g):
    """
    The schema_str is parsed and then all required RAEs are instantiated on the passed graph.
    """
    return graph_generator.parse_schema(g, schema_str)


@func
def schema_str_to_flatgraph(schema_str):
    """
    Generate a Flatgraph containing all the RAEs defining the schema_str.
    """
    from zef.core.VT import FlatGraph
    from zef import  BT
    g = Graph()
    generate_graph_from_file(schema_str, g)
    all_raes = g | now | all | map[lambda zr: origin_rae(zr) if is_a(zr, BT.RELATION_EDGE) else zr] | collect
    return FlatGraph(all_raes)


    # return graph_generator.parse_schema(g, schema_str)



def make_api(schema_root, schema_destination=None, resolvers_destination=None):
    """ This function takes a zefref to the root of the schema graph.
    It then generates the schema file by traversing the graph that is passed.
    From the same graph, the resolver functions for the schema file are generated.
    The schema is then checked for syntax errors and the resolvers are checked for all present
    GQL types in the schema and if this is all successful an API is then created and returned.
    """
    import tempfile
    if schema_destination is None:
        fd,schema_destination = tempfile.mkstemp(suffix=".py", prefix="schema_")
    if resolvers_destination is None:
        resolvers_destination = tempfile.mkdtemp()

    generate_api_schema_file(schema_root, schema_destination)
    generate_resolvers_fcts(schema_root, resolvers_destination)

    resolvers = load_module_from_abs_path("generated_resolvers", os.path.join(resolvers_destination, "ariadne_resolvers.py"))
    schema = load_module_from_abs_path("schema", schema_destination)

    # Need to inject the graph into the resolvers module
    resolvers.g = Graph(schema_root)
    schema = make_executable_schema(schema.generated_schema, resolvers.object_types)
    return schema


triggered_callbacks = []
def create_subscription(payload, schema, g, sng_callback):
    # parsing
    from graphql import parse
    ast = parse(payload['query'])
    name = ast.definitions[0].selection_set.selections[0].name.value

    # register callback
    if name == "onValueChange":
        if 'variables' in payload:
            uid = payload['variables']['uid']
        else:
            # Will throw an error
            uid = ast.definitions[0].selection_set.selections[0].arguments[0].value.value

        def new_value_callback(z):
            response = graphql_sync(schema, payload, context_value={"tx": z | tx, "g": g})
            triggered_callbacks.append((sng_callback, response[1]))

        sub = g[uid] | now | subscribe[on_value_assignment][new_value_callback]
    else:
        sub = None

    return sub

def load_module_from_abs_path(name, path):
    # Why is this so complicated?
    import importlib.util
    spec = importlib.util.spec_from_file_location(name, path)
    foo = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(foo)
    return foo



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

####################################
# * Locating libzef
#----------------------------------

import ctypes, sys, sysconfig, os

# ** Circular import checks
import os
if os.environ.get("ZEFDB_DEVELOPER_CIRCULAR_IMPORTS", "FALSE") == "TRUE":
    from .circular_imports import check_circular_imports
    check_circular_imports()

# ** TLS certificates
# See comments in libzef CMakeLists.txt. This is to pass appropriate certificate
# information to libzef if it is bundled with pyzef.
try:
    # I don't think this will ever fail as python requires an SSL to build now I
    # think. What this block should really be doing is checking for OpenSSL not
    # LibreSSL... or maybe the libzef library should be using LibreSSL on macos.
    import ssl
except ImportError:
    raise ImportError("zef requires OpenSSL to be installed.")

_ssl_paths = ssl.get_default_verify_paths()
os.environ["LIBZEF_CA_BUNDLE"] = _ssl_paths.cafile or ""
os.environ["LIBZEF_CA_PATH"] = _ssl_paths.capath or ""

# ** auth.html location for bundled installs
# This is only relevant for windows, but we can set it regardless
import pathlib
os.environ["LIBZEF_AUTH_HTML_PATH"] = str(pathlib.Path(__file__).resolve().parent.parent)

########################################################
# * Exposing common functions
#------------------------------------------------------
from .. import pyzef
from ._core import *

# Force patching to execute
from . import patching
# Also override merge
from .overrides import *

# This set of imports is to define the order. Later imports are the ones to
# actually provide useful exports.
from . import internals
from . import pure_utils
from . import error
from . import image
from . import op_structs
from . import _ops
from . import VT
from . import zef_functions
from . import abstract_raes
from . import graph_delta
from . import graph_slice
from . import flat_graph
from . import fx
from . import serialization

from .error import Error

from .image import Image

from .fx.fx_types import Effect, FX

from .units import unit

from .graph_slice import GraphSlice


from .flat_graph import FlatGraph_, FlatRef, FlatRefs, Val

# TODO: import the other ValueTypes here and implement constructor by forwarding args
from .VT import (
    TX,
    Nil,
    Any,
    Bool,
    Int,
    Float,
    String,
    Bytes,
    Decimal,
    List,
    Dict,
    Set,
    Stream,
    ValueType,
    Instantiated, 
    Terminated, 
    Assigned,
    FlatGraph,

    Union,
    Intersection,
    Complement,
    Is,
    SetOf,
    RP,
    HasValue,
    Pattern,
    SameAs,

    RelatedOps,
    UsedFor,
    OperatesOn
    )
from .VT.value_type import ValueType_

from .abstract_raes import Entity, AttributeEntity, Relation, TXNode, Root, make_custom_entity

from .zef_functions import func

from .op_structs import ZefOp, LazyValue

from .serialization import serialize, deserialize

from .user_value_type import UserValueType

# instantiating these here, since not all of the core has been
# initialized when Python imports the abstract_raes module
# and a circular import error occurs.
please_instantiate = make_custom_entity(name_to_display='please_instantiate', predetermined_uid='783320c1c3de2610')
please_terminate   = make_custom_entity(name_to_display='please_terminate', predetermined_uid='67cb88b71523f6d9')
please_assign      = make_custom_entity(name_to_display='please_assign',    predetermined_uid='4d4a93522f75ed21')

instantiated     = make_custom_entity(name_to_display='instantiated', predetermined_uid='60252a53a03086b7')
terminated       = make_custom_entity(name_to_display='terminated', predetermined_uid='4f676154ffeb9dc8')
assigned         = make_custom_entity(name_to_display='assigned', predetermined_uid='c31287dab677f38c')

infinity           = make_custom_entity(name_to_display='infinity',    predetermined_uid='4906648460291096')
nil                = make_custom_entity(name_to_display='nil',         predetermined_uid='1654670075329719') #| register_call_handler[f1] | run[execute] | get['entity'] | collect  # TODO


# Implementations come last, so that they can make use of everything else
from . import op_implementations
from .symbolic_expression import SV, SVs, v, unwrap_vars_hack

pyzef.internals.finished_loading_python_core()

# ############################################
# # * Starting the butler
# #------------------------------------------

# # Putting this into a function to avoid poluting namespace
# def _autostart_behaviour():
#     auto_start = ops.config("butler.autoStart", KW.get)
#     if auto_start == "true":
#         internals.initialise_butler()

# _autostart_behaviour()
        



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def c_value_type_check(val, s_typ):
    """Don't call this explicitly. Only for the zefdb core."""

    from ...pyzef.zefops import SerializedValue
    from .._ops import is_a

    typ = s_typ.deserialize()
    if type(val) == SerializedValue:
        val = val.deserialize()

    return is_a(val, typ)

def register_value_type_check():
    from ...pyzef import internals
    internals.register_value_type_check(c_value_type_check)



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .fx_types import Effect
from .._core import EZefRef, ET
from .state import _state
from ..op_structs import Awaitable
from .._ops import *

def stream_create_pushable_stream_handler(eff: Effect):
    """ 
    Create an ET.PushableStream on the local process graph.
    Store the actual Awaitable in fx._state['streams']
    """
    from rx.subject import Subject
    from .. import internals

    g_process = internals.get_local_process_graph()
    # z_stream (an ET.PushableStream) is just a proxy for the actual stream.
    # Both on push and subscribing / applying further transformations
    z_stream = to_ezefref(ET.PushableStream | g_process | run)
    _state['streams'][z_stream] = Subject()
    # return { 'stream': my_stream }    # too verbose?
    return Awaitable(z_stream, True)


def stream_push_handler(eff: Effect):   # -> Union[Nil, Error]
    assert 'stream' in eff
    s = eff['stream']
    z_stream = s.stream_ezefref
    assert isinstance(s, Awaitable)
    streams = _state['streams']
    if not z_stream in streams:
        raise RuntimeError(f'in push effect handler: associated stream not found in fx._state["streams"]')
    
    # TODO: check that type of item agrees if stream is typed!!!!!!!!!!!!!!!!!!
    streams[z_stream].on_next(eff['item'])
    return None


def stream_complete_handler(eff: Effect):
    raise NotImplementedError()




# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..core import *
from ..ops import *
from ariadne import ObjectType,MutationType,QueryType,InterfaceType,SubscriptionType


#--------------------------Schema Generator-------------------------
def generate_schema_str(schema_dict: dict) -> str:
    """
    Generate a GraphQL schema string from a GraphQL schema dict.
    """

    def parse_enums(enums_list):
        def parse_enum(tup):
            enum, options = tup
            return (f"enum {enum}" + "{\n" 
                    + "\n".join(f"    {option}" for option in options.keys())
                    + "\n}"
            )
        
        return (
            enums_list
            | items
            | map[parse_enum]
            | join["\n"]
            | collect
        )

    def parse_scalars(scalars):
        return (
            scalars
            | items
            | map[lambda x: f"scalar {x[0]}"]
            | join["\n"]
            | collect
        )

    @func
    def parse_interfaces_or_types(d, prefix):
        def parse_implements(interfaces):
            if interfaces: return f"implements {''.join(interfaces)} "
            return ""

        def parse_interface_or_type(tup):
            name, fields_d = tup
            return (
                f"{prefix} {name} {parse_implements(fields_d.get('_interfaces', []))}" + "{\n"
                + parse_fields(fields_d)
                + "\n}"
            )

        return (
            d
            | items
            | map[parse_interface_or_type]
            | join["\n\n"]
            | collect
        )
    
    
    
    def parse_args(args):
        def parse_arg(tup):
            def parse_default_arg():
                if "default" in arg_dict: return f" = {arg_dict['default']}"
                return ""
            arg_name, arg_dict = tup
            return f"{arg_name}: {arg_dict['type']}{parse_default_arg()}"

        if args:
            args = (
            args
            | items
            | map[parse_arg]
            | join[", "]
            | collect
            )
            return f"({args})"
        
        return ""


    def parse_fields(fields_d):
        def parse_field(tup):
            field_name, field_dict = tup
            if field_name.startswith("_"): return ""
            return f"  {field_name}{parse_args(field_dict.get('args', []))}: {field_dict['type']}" 

        return (
            fields_d
            | items
            | map[parse_field]
            | join["\n"]
            | collect
        )
    
    def parse_inputs(input_d):
        def parse_input_field(fields_d):
            return (
                fields_d
                | items
                | map[lambda x: f"    {x[0]}: {x[1]}"]
                | join["\n"]
                | collect
            )
        def parse_input_type(tup):
            name, fields_d = tup
            return (
                f"input {name}" + "{\n"
                + parse_input_field(fields_d)
                + "\n}"
            )
        return (
            input_d
            | items
            | map[parse_input_type]
            | join["\n\n"]
            | collect
        )


    allowed_keys = ["_Interfaces", "_Types", "_Scalars", "_Enums", "_Inputs"]
    schema_dict = select_keys(schema_dict, *allowed_keys)
    dispatch = {
        "_Interfaces":  parse_interfaces_or_types["interface"],
        "_Types":       parse_interfaces_or_types["type"],
        "_Scalars":     parse_scalars,
        "_Enums":       parse_enums,
        "_Inputs":      parse_inputs,
    }

    return (
        schema_dict
        | items
        | map[lambda x: dispatch[x[0]](x[1])]
        | join["\n"]
        | collect
    )



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..pyzef.main import *
from ..pyzef import main, zefops, internals
from ._core import ET, AttributeEntityType, EntityType

main.ZefRef.__hash__ = lambda self: hash(((index(self)), index(self | zefops.tx)))
main.EZefRef.__hash__ = lambda self: index(self)

main.QuantityFloat.__hash__ = lambda self: hash((self.value, self.unit))
main.QuantityInt.__hash__ = lambda self: hash((self.value, self.unit))
main.Time.__hash__ = lambda self: hash(self.seconds_since_1970)

internals.Delegate.__hash__ = lambda self: hash((self.order, self.item))
internals.DelegateRelationTriple.__hash__ = lambda self: hash((self.rt, self.source, self.target))
internals.DelegateTX.__hash__ = lambda self: hash(internals.DelegateTX)
internals.DelegateRoot.__hash__ = lambda self: hash(internals.DelegateRoot)


from ..pyzef.main import EntityType, RelationType
from ..pyzef.internals import AttributeEntityType, EntityTypeStruct, RelationTypeStruct, AttributeEntityTypeStruct, BlobTypeStruct, BlobType
override_types = (EntityTypeStruct, RelationTypeStruct, AttributeEntityTypeStruct, BlobTypeStruct, BlobType, AttributeEntityType, EntityType, RelationType)

def or_for_types(self, other):
    from . import ValueType_
    allowed_types = (ValueType_,) + override_types
    if isinstance(other, allowed_types):
        return ValueType_(type_name='Union', absorbed=(self, other,))

    return NotImplemented
    
def and_for_types(self, other):
    from . import ValueType_
    allowed_types = (ValueType_,) + override_types
    if isinstance(other, allowed_types):
        return ValueType_(type_name='Intersection', absorbed=(self, other,))

    return NotImplemented

for x in override_types:
    x.__or__ = or_for_types
    x.__ror__ = lambda x,y: or_for_types(y,x)
    x.__and__ = and_for_types
    x.__rand__ = lambda x,y: and_for_types(y,x)


# FIXME: why do we need the standard list of timezones in the docstring for Time? Should just refer people to the TZ database. I have removed it for now.

# ------------------------- patch Time constructor to allow create of zef Time objects both from time stamps and readable strings ------------------------
from datetime import datetime, timezone, timedelta
import dateparser
import pytz   # for time zones

def create_monkey_patched_Time_ctor(old_init_fct):
    def monkey_patched_Time_ctor(self, x, timezone: str = 'Asia/Singapore'):
        """ x could be a time stamp (float/int) or a string to be parsed
        Note: the python library date parser parses dates of the following form UNINTUITIVELY, i.e. it assumes the MDY convention:  
            '9.2.2015 13:59' is read as Sep. 2. 2015
            Hence: ALWAYS stick to the convention of transforming strings to the form Time('2019-07-15 13:59', timezone='Europe/London')

            Other formats work, e.g.
            Time('9. Feb 2015 13:59')
            Time('Martes 21 de Octubre de 2014 13:23:12')    # various languages
            Time('1 เดือนตุลาคม 2005, 1:00 AM')    # Thai
            Time('1 hour ago')         # relative dates
            Time('now')

        The time zone can be specified as a string: 
            t0 = Time('2019-07-15 13:59', timezone='Europe/Berlin')
        If no time zone is specified, Singapore is assumed as a default for now - this will be cleaned up in the future.
        

        Valid timezones: see TZ database.
        """
        if(isinstance(x, float)):
            self = old_init_fct(self, x)
        elif(isinstance(x, int)):
            self = old_init_fct(self, float(x))
        elif(isinstance(x, zefops.Now)):
            self = old_init_fct(self, x)
        elif(isinstance(x, str)):
            try:
                my_timestamp = datetime.timestamp(dateparser.parse(x, settings={'TIMEZONE': timezone, 'RETURN_AS_TIMEZONE_AWARE': True}))
                self = old_init_fct(self, my_timestamp)
            except:
                raise ValueError(f"Could not parse the following string as a valid datetime: \"{x}\"")
        else:
            raise ValueError('Invalid argument type passed to Time constructor')
    return monkey_patched_Time_ctor
main.Time.__init__ = create_monkey_patched_Time_ctor(Time.__init__)


def _to_date_time(self: Time, timezone:str='Asia/Singapore', str_format: str='%Y-%m-%d %H:%M:%S'):
    return datetime.fromtimestamp(self.seconds_since_1970)\
            .astimezone(pytz.timezone(timezone))\
            .strftime(str_format)

def _to_time(self: Time, timezone:str='Asia/Singapore'):
    return _to_date_time(self, timezone, str_format='%H:%M:%S')

def _to_date(self: Time, timezone:str='Asia/Singapore'):
    return _to_date_time(self, timezone, str_format='%Y-%m-%d')



main.Time.date_time = _to_date_time
main.Time.time = _to_time
main.Time.date = _to_date


def _repr_for_zef_time(self):
    return datetime.fromtimestamp(self.seconds_since_1970)\
                .astimezone(timezone(timedelta(hours=8)))\
                .strftime(f"<Time %Y-%m-%d %H:%M:%S (+0800)>")
def _str_for_zef_time(self):
    return datetime.fromtimestamp(self.seconds_since_1970)\
                .astimezone(timezone(timedelta(hours=8)))\
                .strftime(f"%Y-%m-%d %H:%M:%S (+0800)")
main.Time.__repr__ = _repr_for_zef_time
main.Time.__str__ = _str_for_zef_time


def help(arg):
    import builtins
    from . import ZefRef
    from ..zefops.base import L, value
    if isinstance(arg, ZefRef) and arg | BT == BT.ENTITY_NODE and arg | ET == ET.ZEF_Function:
        if len(arg >> L[RT.DocString]) == 1:
            print((arg >> RT.DocString) | value.String)
        else:
            print("This ZEF_Function doesn't have any DocString")
    else:
        print(builtins.help(arg))



# Temp fix fow now: these operators should be overloaded in C++ and return a new C++ type.
def _rae_get_item(self, n):
    """
    we want to write  RT.UsedBy['some internal id']. 
    Same for ETs, AETs. Monkeypatch that from Python for now.
    """
    from ._ops import instantiated
    # return {'RAE': self, 'capture': n}
    return instantiated[self][n]




main.QuantityFloat.__getitem__ = _rae_get_item
main.QuantityInt.__getitem__ = _rae_get_item



# ---------------------------- special treatment for EntityType ------------------------------------

def absorbed_get_item(self, x):
    # don't mutate, create a new instance
    from copy import copy
    new_obj = copy(self)
    if '_absorbed' not in self.__dict__:
        new_obj._absorbed = (x, )
    else:
        new_obj._absorbed = (*self._absorbed, x)
    return new_obj

def eq_with_absorbed(x, y, orig_eq):
    orig_res = orig_eq(x, y)
    if orig_res == NotImplemented:
        return NotImplemented
    
    if orig_res == False:
        return False

    try:
        absrbd1 = x.__getattribute__('_absorbed')
    except AttributeError:
        absrbd1 = ()
    
    try:
        absrbd2 = y.__getattribute__('_absorbed')
    except AttributeError:
        absrbd2 = ()

    return absrbd1 == absrbd2

def wrap_eq(typ):
    orig = typ.__eq__
    typ.__eq__ = lambda x,y,orig=orig: eq_with_absorbed(x, y, orig)
    import types
    assert type(typ.__ne__) == types.WrapperDescriptorType

def hash_with_absorbed(self, orig_hash):
    orig_res = orig_hash(self)
    try:
        absrbd = self.__getattribute__('_absorbed')
    except AttributeError:
        absrbd = ()
    return hash((orig_res,) + absrbd)

def wrap_hash(typ):
    orig = typ.__hash__
    typ.__hash__ = lambda self,orig=orig: hash_with_absorbed(self, orig)
    

def repr_with_absorbed(self, orig_repr):
    original = orig_repr(self)
    if '_absorbed' not in self.__dict__:
        return original
    else:
        return original + ''.join(('[' + repr(el) + ']' for el in self._absorbed))

def wrap_repr(typ):
    orig = typ.__repr__
    typ.__repr__ = lambda self,orig=orig: repr_with_absorbed(self, orig)
    


main.EntityType.__getitem__ = absorbed_get_item
wrap_repr(main.EntityType)
wrap_eq(main.EntityType)
wrap_hash(main.EntityType)

main.RelationType.__getitem__ = absorbed_get_item
wrap_repr(main.RelationType)
wrap_eq(main.RelationType)
wrap_hash(main.RelationType)

main.Keyword.__getitem__ = absorbed_get_item
wrap_repr(main.Keyword)
wrap_eq(main.Keyword)
wrap_hash(main.Keyword)

internals.AttributeEntityType.__getitem__ = absorbed_get_item
# wrap_repr(internals.AttributeEntityType)
wrap_eq(internals.AttributeEntityType)
wrap_hash(internals.AttributeEntityType)

internals.Delegate.__getitem__ = absorbed_get_item
wrap_repr(internals.Delegate)
wrap_eq(internals.Delegate)
wrap_hash(internals.Delegate)






# we want to allow for e.g. "AET.Float <= 42.1"
# return a ZefOp to use as a container here.
# We used a dict previously, but then we could not 
# catch the case "AET.Float['abc'] <= 42.1"
# For consistency, we use Lazy ZefOps as intermediate containers
def leq_monkey_patching_ae(self, other):
    from ._ops import assign, LazyValue
    return LazyValue(self) | assign[other]
    
AttributeEntityType.__le__ = leq_monkey_patching_ae


# # Pretty printing for ZefRefs
# def pprint_ZefRefs(self, p, cycle):
#     p.text(str(self))
#     p.text(" [\n")

#     N_max = 10
#     if len(self) > N_max:
#         # Split into first part and last part
#         N = N_max // 2
#         first = [self[i] for i in range(N)]
#         last = [self[i] for i in range(len(self)-N, len(self)-1)]
#         p.text('\n'.join("\t"+str(x) for x in first))
#         p.text('\n\t...\n')
#         p.text('\n'.join("\t"+str(x) for x in last))
#     else:
#         p.text('\n'.join("\t"+str(x) for x in self))
#     p.text(']')

# main.ZefRefs._repr_pretty_ = pprint_ZefRefs
# main.EZefRefs._repr_pretty_ = pprint_ZefRefs

def convert_to_assign(self, value):
    from ._ops import assign
    return self | assign[value]
ZefRef.__le__ = convert_to_assign
EZefRef.__le__ = convert_to_assign

original_Graph__contains__ = main.Graph.__contains__
def Graph__contains__(self, x):
    from .abstract_raes import Entity, AttributeEntity, Relation
    from ._ops import origin_uid
    if type(x) in [Entity, AttributeEntity, Relation]:
        return origin_uid(x) in self

    if type(x) in [ZefRef, EZefRef]:
        if Graph(x) == self:
            return True
        return origin_uid(x) in self

    return original_Graph__contains__(self, x)
main.Graph.__contains__ = Graph__contains__
    
original_Graph__getitem__ = main.Graph.__getitem__
def Graph__getitem__(self, x):
    from .abstract_raes import Entity, AttributeEntity, Relation
    from ._ops import uid, target
    from .internals import BT
    if type(x) in [Entity, AttributeEntity, Relation]:
        return self[uid(x)]

    res = original_Graph__getitem__(self, x)
    # We magically transform any FOREIGN_ENTITY_NODE accesses to the real RAEs.
    # Accessing the low-level BTs can only be done through traversals
    if BT(res) in [BT.FOREIGN_ENTITY_NODE, BT.FOREIGN_ATTRIBUTE_ENTITY_NODE, BT.FOREIGN_RELATION_EDGE]:
        # return target(res << BT.ORIGIN_RAE_EDGE)
        # TODO: We need a consistent way of accessing this.
        #
        # Is this even possible for a EZefRef? User can terminate and recreate
        # RAEs that have the same identity, so there's nothing sensible that
        # could be returned here?
        #
        # Hence, going to disable this. It will be that you can only get the RAE
        # corresponding to a foreign RAE in the context of a graph slice. Doing
        # it here will just return the low level blob.
        pass
    return res

main.Graph.__getitem__ = Graph__getitem__

original_Graph__init__ = main.Graph.__init__
def Graph__init__(self, *args, **kwds):
    from .graph_slice import GraphSlice
    if len(kwds) == 0 and len(args) == 1 and isinstance(args[0], GraphSlice):
        return original_Graph__init__(self, args[0].tx)

    return original_Graph__init__(self, *args, **kwds)
main.Graph.__init__ = Graph__init__


from ..pyzef.zefops import SerializedValue

def SerializedValue_deserialize(self):
    if self.type == "tools.serialize":
        from .serialization import deserialize
        from json import loads
        return deserialize(loads(self.data))
    else:
        print(self.type)
        print(self.data)
        raise Exception(f"Don't know how to deserialize a type of {self.type}")
SerializedValue.deserialize = SerializedValue_deserialize

def SerializedValue_serialize(value):
    from .serialization import serialize, serialization_mapping, is_python_scalar_type
    from json import dumps
    if type(value) in serialization_mapping or is_python_scalar_type(value):
        return SerializedValue("tools.serialize", dumps(serialize(value)))
    else:
        raise Exception(f"Don't know how to serialize a type of {value}")
SerializedValue.serialize = SerializedValue_serialize

def SerializedValue_repr(self):
    return f"SerializedValue('{self.type}', '{self.data}')"
SerializedValue.__repr__ = SerializedValue_repr



def AttributeEntityTypeStruct_getitem(self, x):
    return AttributeEntityType(SerializedValue.serialize(x))
AttributeEntityTypeStruct.__getitem__ = AttributeEntityTypeStruct_getitem

original_AttributeEntityType__repr__ = internals.AttributeEntityType.__repr__
def AttributeEntityType_repr(self):
    s = "AET"
    if self.complex_value:
        s += "(" + str(self.complex_value.deserialize()) + ")"
    else:
        s += "." + str(self.rep_type)
    if '_absorbed' in self.__dict__:
        s += ''.join(('[' + repr(el) + ']' for el in self._absorbed))
    return s
AttributeEntityType.__repr__ = AttributeEntityType_repr








#                           _____         _    _  _               ___   _        _              _       _   _         _           _    _                                       
#                          | ____| _ __  | |_ (_)| |_  _   _     / _ \ | |__    (_)  ___   ___ | |_    | \ | |  ___  | |_   __ _ | |_ (_)  ___   _ __                          
#   _____  _____  _____    |  _|  | '_ \ | __|| || __|| | | |   | | | || '_ \   | | / _ \ / __|| __|   |  \| | / _ \ | __| / _` || __|| | / _ \ | '_ \     _____  _____  _____ 
#  |_____||_____||_____|   | |___ | | | || |_ | || |_ | |_| |   | |_| || |_) |  | ||  __/| (__ | |_    | |\  || (_) || |_ | (_| || |_ | || (_) || | | |   |_____||_____||_____|
#                          |_____||_| |_| \__||_| \__| \__, |    \___/ |_.__/  _/ | \___| \___| \__|   |_| \_| \___/  \__| \__,_| \__||_| \___/ |_| |_|                        
#                                                      |___/                  |__/                                                                                             


class EntityValueInstance_:
    def __init__(self, entity_type, **kwargs):
        self._entity_type: EntityType = entity_type
        self._kwargs = kwargs
        
    def __repr__(self):
        nl = '\n'
        return f'ET.{str(self._entity_type)}({f", ".join([f"{k}={repr(v)}" for k, v in self._kwargs.items()])})'
    
    def __getattr__(self, name):
        return self._kwargs[name]




def entity_type_call_func(self, *args, **kwargs):
    return EntityValueInstance_(EntityType(self.value), **kwargs)

EntityType.__call__ = entity_type_call_func





# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class Bytes_:
    """
    The default repr uses hexadecimal encoding.
    """
    def __init__(self, x):
        import string
        if isinstance(x, bytes):
            self.data = x
        elif isinstance(x, str) and all(c in string.hexdigits for c in x):
            self.data = x.encode()    # use utf8
        else:
            raise NotImplementedError(f"constructing a bytes from x={x}")
    
    def __repr__(self):
        return f'Bytes("{self.data.hex()}")'

    def __str__(self):
        return self.data.hex()

    def __bytes__(self):
        """
        allow casting to python bytes
        >>> bytes(my_zef_bytes)
        """
        return self.data




# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    "config"
]

import os
import yaml
from ..core.op_implementations.dispatch_dictionary import _op_to_functions
from ..core import *
from ..core._ops import *
from ..pyzef import main as pymain

def get_config(key):
    return pymain.get_config_var(key)

def set_config(key, val):
    return pymain.set_config_var(key, val)

def list_config(filter):
    tuples = pymain.list_config(filter)

    d = {}
    for path,val in tuples:
        d = insert_in(d, path.split('.'), val)
    return d


def config_implementation(payload, action):
    if action == KW.set:
        assert isinstance(payload, tuple)
        assert len(payload) == 2
        assert isinstance(payload[0], str)
        set_config(payload[0], payload[1])
    elif action == KW.get:
        assert isinstance(payload, str)
        return get_config(payload)
    elif action == KW.list:
        assert isinstance(payload, str)
        return list_config(payload)
    else:
        raise NotImplementedError("Action {} not implemented".format(action))


def config_typeinfo(v_tp):
    return VT.Any


_op_to_functions[RT.Config] = (config_implementation, config_typeinfo)

config = ZefOp(((RT.Config, ()), ))



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import *
from .._ops import *
from ..VT import Int, Float, String, Bool, Nil, List, Dict


def handle_int(x: Int, name: str):
    return [AET.Int[name] <= x]


def handle_float(x: Float, name: str):
    return [AET.Float[name] <= x]


def handle_string(x: String, name: str):
    return [AET.String[name] <= x]


def handle_bool(x: Bool, name: str):
    return [AET.Bool[name] <= x]


def handle_nil(x: Nil, name: str):
    return [ET.Nil[name]]


def handle_dict(dd: dict, name: str):    

    def generate_random_name() -> str:
        from random import randint
        return f"_{randint(0, 1000000000)}"

    def make_delta(rt_and_trg):
        rel_name, targ = rt_and_trg
        rt = RT(to_pascal_case(rel_name))
        new_name = generate_random_name()
        new_delta = to_deltas(targ, new_name)
        return [
            *new_delta,
            (Z[name], rt, Z[new_name]),
        ]

    try:
        type_name = to_pascal_case(dd['_type'])     
    except KeyError:
        type_name = 'ZEF_Unspecified'

    return [
    ET(type_name)[name],
    *(dd 
     | items 
     | filter[lambda p: p[0] != '_type']
     | map[make_delta]
     | concat     
     | collect
    )
    ]


def handle_list(x: List, name: str):
    def generate_random_name() -> str:
        from random import randint
        return f"_{randint(0, 100000000)}"

    list_edge_count = 0
    ed_name = lambda m: f"ed_{name}_{m}"
    def make_delta_sublist(list_el):
        new_name = generate_random_name()
        new_delta = to_deltas(list_el, new_name)
        nonlocal list_edge_count
        list_edge_count += 1
        return [
            *new_delta,
            (Z[name], RT.ZEF_ListElement[ed_name(list_edge_count-1)], Z[new_name]),
        ]

    res = [
        ET.ZEF_List[name],
        *(x
          | map[make_delta_sublist]
          | concat
          | collect
        )
    ]
    list_order_rels = (
        range(list_edge_count)
        | sliding[2]
        | map[lambda p: (Z[ed_name(p[0])], RT.ZEF_NextElement, Z[ed_name(p[1])])]
        | collect
    )
    return [*res, *list_order_rels]




def to_deltas(x, name: str) -> List:
    d_dispatch = {
        int: handle_int,
        float: handle_float,
        str: handle_string,
        bool: handle_bool,
        type(None): handle_nil,
        dict: handle_dict,
        list: handle_list,
        tuple: handle_list,
    }
    return d_dispatch[type(x)](x, name=name)



def to_flatgraph_imp(d: dict):
    """
    An operator that allows transforming plain Zef values (dicts, lists, 
    scalars and nested forms thereof) to FlatGraphs in a standardized form.

    Dicts: if a key "_type" is specified, this is not attached as child,
    but the associated value is assumed to be a string, which is used 
    as the type name. All other keys are attached as children.

    Lists: These are converted to ZefLists, and the elements are attached.

    ---- Examples -----
    >>> d = {
    ... '_type': 'paragraph',
    ... 'foo': 42,
    ... 'bar': [1, 2, 'bye', False]
    ... }
    ... d | to_flatgraph     # returns a FlatGraph
        
    ---- Signature ----
    (Bool | Int | Float | String | Dict | List) -> FlatGraph

    ---- Tags ----
    - operates on: List
    - operates on: Dict

    """
    fg = FlatGraph(
        to_deltas(d, name='root')
    )
    # keep this function pure
    # we don't want the random names leaking out
    fg.key_dict = select_keys(fg.key_dict, 'root')    
    return fg




# Copyright (c) 2022 Synchronous Technologies Pte Ltd
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from ..core import *
from ..ops import *
from graphql import parse
from graphql.language.ast import (
    ScalarTypeDefinitionNode,
    ObjectTypeDefinitionNode, NamedTypeNode,
    ListTypeNode, NonNullTypeNode,
    InterfaceTypeDefinitionNode,
    InputObjectTypeDefinitionNode,
    EnumTypeDefinitionNode,
)

# TODO add support for parsing enums

def generate_schema_dict(schema_str: str) -> dict:
    """
    Given a GraphQL schema string, generate a dictionary representation of the schema.
    """
    def resolve_interfaces(interfaces):
        if interfaces: return {'_interfaces': interfaces | map[lambda i: i.name.value] | collect}
        return {}

    def resolve_args(args):
        def resolve_arg(arg):
            def resolve_default_value(arg):
                if arg.default_value:
                    return {"default": arg.default_value.value}
                return {}

            return {
                arg.name.value: {
                    "type": resolve_type(arg.type),
                    **resolve_default_value(arg),
                }
            }
        
        if args: return {'args': args | map[resolve_arg] | merge | collect}
        return {}

    def resolve_type(type):
        return LazyValue(type) | match[
            (Is[lambda t: isinstance(t, NamedTypeNode)], lambda t: t.name.value),
            (Is[lambda t: isinstance(t, ListTypeNode)], lambda t: f"[{resolve_type(t.type)}]"),
            (Is[lambda t: isinstance(t, NonNullTypeNode)], lambda t: f"{resolve_type(t.type)}!"),
        ] | collect

    def resolve_field(field):
        return {
            field.name.value: {
                'type': resolve_type(field.type),
                'resolver': None,
                **resolve_args(list(field.arguments)),
            }
        }

    @func
    def dispatch_on_type_or_interface(type_node, is_interface = False):
        interface_resolver = {"_interface_resolver": None} if is_interface else {}
        return {
            type_node.name.value : {
                **(list(type_node.fields) | map[resolve_field] | merge | collect),
                **resolve_interfaces(list(type_node.interfaces)),
                **interface_resolver,
            }
        }

    def dispatch_on_scalar(type_node):
        return {
            type_node.name.value: {
                'parser': None,
                'serializer': None,
            }
        } 
    
    def dispatch_on_input(input_node):
        def resolve_input_field(field):
            return { field.name.value: resolve_type(field.type) } 
        return {
            input_node.name.value: {
                **(list(input_node.fields) | map[resolve_input_field] | merge | collect),
            }
        } 

    def dispatch_on_enum(enum_node):
        return {
            enum_node.name.value: {
                **(list(enum_node.values) | map[lambda v: {v.name.value:v.name.value}] | merge | collect),
            }
        } 

    
    dispatch = match[
        (Is[lambda t: t[0] == ObjectTypeDefinitionNode], lambda g: {"_Types": g[1] | map[dispatch_on_type_or_interface] | merge | collect}),
        (Is[lambda t: t[0] == InterfaceTypeDefinitionNode], lambda g: {"_Interfaces": g[1] | map[dispatch_on_type_or_interface[True]] | merge | collect}),
        (Is[lambda t: t[0] == ScalarTypeDefinitionNode], lambda g: {"_Scalars": g[1] | map[dispatch_on_scalar] | merge | collect}),
        (Is[lambda t: t[0] == InputObjectTypeDefinitionNode], lambda g: {"_Inputs": g[1] | map[dispatch_on_input] | merge | collect}),
        (Is[lambda t: t[0] == EnumTypeDefinitionNode], lambda g: {"_Enums": g[1] | map[dispatch_on_enum] | merge | collect}),
    ]

    document = parse(schema_str, no_location=True)
    return (
        list(document.definitions)
        | group_by[type]
        | map[dispatch]
        | merge
        | collect
    )


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Assuming this file is not imported during zefdb init.
from ... import *
from ...ops import *
from functools import partial as P
from ...core.logger import log
import functools

from ariadne import ObjectType, QueryType, MutationType, EnumType, ScalarType

##############################
# * Utils
#----------------------------

@func
def is_core_scalar(z):
    return (
        is_a(z, AET.String) or
        is_a(z, AET.Int) or
        is_a(z, AET.Float) or
        is_a(z, AET.Bool)
    )
    
assert_type = Assert[is_a[ET.GQL_Type | ET.GQL_Enum | AET]][lambda z: f"{z} is not a GQL type"]
assert_field = Assert[is_a[RT.GQL_Field]][lambda z: f"{z} is not a GQL field"]

optional = single_or[None]

@func
def OutO(z, rt):
    if z is None:
        return None
    return z | Outs[rt] | optional | collect

@func
def fvalue(z, rt, *args):
    if len(args) == 0:
        return z | Out[rt] | value | collect
    else:
        default, = args
        return z | OutO[rt] | value_or[default] | collect

        
op_is_scalar = assert_type | is_a[AET | ET.GQL_Enum]
op_is_orderable = assert_type | is_a[AET.Float | AET.Int | AET.Time]
op_is_summable = assert_type | is_a[AET.Float | AET.Int]
op_is_stringlike = assert_type | is_a[AET.String]
op_is_list = assert_field | fvalue[RT.List][False] | collect
op_is_required = assert_field | fvalue[RT.Required][False] | collect
op_is_unique = assert_field | fvalue[RT.Unique][False] | collect
op_is_searchable = assert_field | fvalue[RT.Search][False] | collect
op_is_aggregable = assert_field | And[Not[op_is_list]][target | Or[op_is_orderable][op_is_summable]]
op_is_incoming = assert_field | fvalue[RT.Incoming][False] | collect

op_is_upfetch = assert_field | fvalue[RT.Upfetch][False] | collect
op_upfetch_field = (assert_type | out_rels[RT.GQL_Field]
                    | filter[op_is_upfetch]
                    | optional
                    | collect)
op_has_upfetch = op_upfetch_field | Not[equals[None]] | collect

########################################
# * Generating it all
#--------------------------------------

def generate_resolvers_fcts(schema_root):
    query_dict = {}
    mutation_dict = {}

    types_dict = {"Query": query_dict,
                  "Mutation": mutation_dict}
    input_types_dict = {}
    enums_dict = {}
    scalars_dict = {}

    full_dict = {"_Types": types_dict,
                 "_Inputs": input_types_dict,
                 "_Enums": enums_dict,
                 "_Scalars": scalars_dict}

    scalars_dict["DateTime"] = {
        "serializer": apply[lambda t: datetime.datetime.fromtimestamp(t.seconds_since_1970).isoformat()],
        "parser": apply[Time]
    }


    for z_type in schema_root | Outs[RT.GQL_Type]:
        name = z_type | F.Name | collect

        lower_name = to_camel_case(name)
        has_upfetch = op_has_upfetch(z_type)

        ref_name = name + "Ref"
        add_input_name = "Add" + name + "Input"
        update_input_name = "Update" + name + "Input"
        filter_name = name + "Filter"
        mutate_response_name = "Mutate" + name + "Response"
        aggregate_response_name = "Aggregate" + name + "Response"

        type_dict = {"id": {
            "type": "ID!", "resolver": resolve_id2
        }}
        ref_input_dict = {"id": "ID"}
        add_input_dict = {"id": "ID"}
        aggregate_fields_dict = {
            "count": {"type": "Int", "resolver": get["count"]},
        }
        if has_upfetch:
            # Note: upfetch does not get an id field
            upfetch_input_dict = {}
            upfetch_input_name = "Upfetch" + name + "Input"
            
        for z_field in z_type | out_rels[RT.GQL_Field]:
            field_name = z_field | F.Name | collect

            is_scalar = z_field | target | op_is_scalar | collect
            is_required = z_field | op_is_required | collect
            is_list = z_field | op_is_list | collect
            is_orderable = z_field | target | op_is_orderable | collect
            is_summable = z_field | target | op_is_summable | collect
            is_aggregable = z_field | op_is_aggregable | collect

            if op_is_unique(z_field):
                assert is_scalar, "Unique fields must be scalars"
                assert not is_list, "Unique fields can't be lists"

            if op_is_upfetch(z_field):
                assert op_is_unique(z_field), "Upfetch field must be unique"
                assert is_required, "Upfetch field must be required"

            base_field_type = z_field | target | F.Name | collect
            if is_scalar:
                base_ref_field_type = base_field_type
            else:
                base_ref_field_type = base_field_type + "Ref" 

            if is_required:
                maybe_required = "!"
            else:
                maybe_required = ""
            if is_list:
                maybe_params = schema_generate_list_params(target(z_field), full_dict)
                ref_field_type = "[" + base_ref_field_type + "]"
                add_field_type = ref_field_type
                upfetch_field_type = "[" + base_ref_field_type + maybe_required + "]" + maybe_required
                field_type = "[" + base_field_type + maybe_required + "]" + maybe_required
            else:
                maybe_params = {}
                ref_field_type = base_ref_field_type
                add_field_type = ref_field_type
                upfetch_field_type = base_ref_field_type + maybe_required
                field_type = base_field_type + maybe_required

            type_dict[field_name] = {
                "type": field_type,
                # "resolver": P(resolve_field, z_field=z_field),
                "resolver": resolve_field2[z_field],
                "args": maybe_params,
            }

            add_input_dict[field_name] = add_field_type
            ref_input_dict[field_name] = ref_field_type
            if has_upfetch:
                upfetch_input_dict[field_name] = upfetch_field_type

            if is_aggregable:
                if is_orderable:
                    aggregate_fields_dict[f"{field_name}Min"] = {
                        "type": ref_field_type,
                        "resolver": get[f"{field_name}Min"],
                    }
                    aggregate_fields_dict[f"{field_name}Max"] = {
                        "type": ref_field_type,
                        "resolver": get[f"{field_name}Max"],
                    }
                if is_summable:
                    aggregate_fields_dict[f"{field_name}Sum"] = {
                        "type": ref_field_type,
                        "resolver": get[f"{field_name}Sum"],
                    }
                    aggregate_fields_dict[f"{field_name}Avg"] = {
                        "type": ref_field_type,
                        "resolver": get[f"{field_name}Avg"],
                    }

        query_params = schema_generate_list_params(z_type, full_dict)

        types_dict[name] = type_dict
        input_types_dict[ref_name] = ref_input_dict
        input_types_dict[add_input_name] = add_input_dict
        input_types_dict[update_input_name] = {
            "filter": filter_name + "!",
            "set": ref_name,
            "remove": ref_name,
        }
        types_dict[mutate_response_name] = {
            "count": {"type": "Int", "resolver": get["count"]},
            lower_name: {
                "type": f"[{name}]",
                "args": query_params,
                # "resolver": apply[P(resolve_filter_response, type_node=z_type)],
                "resolver": resolve_filter_response2[z_type],
            }
        }
        types_dict[aggregate_response_name] = aggregate_fields_dict

        if has_upfetch:
            input_types_dict[upfetch_input_name] = upfetch_input_dict

        # Add the 3 top-level queries
        query_dict[f"get{name}"] = {
            "type": name,
            "args": {"id": {"type": "ID!"}},
            # "resolver": apply[P(resolve_get, type_node=z_type)],
            "resolver": resolve_get2[z_type],
        }
        query_dict[f"query{name}"] = {
            "type": f"[{name}]",
            "args": query_params,
            # "resolver": apply[P(resolve_query, type_node=z_type)],
            "resolver": resolve_query2[z_type],
        }
        query_dict[f"aggregate{name}"] = {
            "type": aggregate_response_name,
            "args": query_params,
            # "resolver": apply[P(resolve_aggregate, type_node=z_type)],
            "resolver": resolve_aggregate2[z_type],
        }

        # Add the 3 top-level mutations
        mutation_dict[f"add{name}"] = {
            "type": mutate_response_name,
            "args": {
                "input": {"type": f"[{add_input_name}!]!"},
                "upsert": {"type": "Boolean"}
            },
            # "resolver": apply[P(resolve_add, type_node=z_type)],
            "resolver": resolve_add2[z_type],
        }
        mutation_dict[f"update{name}"] = {
            "type": mutate_response_name,
            "args": {"input": {"type": f"{update_input_name}!"}},
            # "resolver": apply[P(resolve_update, type_node=z_type)],
            "resolver": resolve_update2[z_type],
        }
        mutation_dict[f"delete{name}"] = {
            "type": mutate_response_name,
            "args": {"filter": {"type": f"{filter_name}!"}},
            # "resolver": apply[P(resolve_delete, type_node=z_type)],
            "resolver": resolve_delete2[z_type],
        }
        if has_upfetch:
            mutation_dict[f"upfetch{name}"] = {
                "type": mutate_response_name,
                "args": {"input": {"type": f"[{upfetch_input_name}!]!"}},
                # "resolver": apply[P(resolve_upfetch, type_node=z_type)],
                "resolver": resolve_upfetch2[z_type],
            }

    for z_enum in schema_root | Outs[RT.GQL_Enum]:
        name = z_enum | F.Name | collect

        opts = {}
        for z_opt in z_enum | Outs[RT.GQL_Field]:
            assert is_a(z_opt, AET.Enum(name))
            opt_en = value(z_opt)
            opts[opt_en.enum_value] = opt_en

        enums_dict[name] = opts

    # Always generate the Int scalar type 
    int_type = schema_root | Outs[RT.GQL_CoreScalarType][AET.Int] | single | collect
    schema_generate_type_dispatch(int_type, full_dict)

    return full_dict

################################################
# * Schema specific parts
#----------------------------------------------

def schema_generate_type_dispatch(z_type, full_dict):
    filter_name = (z_type | F.Name | collect) + "Filter"
    if filter_name not in full_dict["_Inputs"]:
        full_dict["_Inputs"][filter_name] = None
        # These will update the full_dict automatically
        if op_is_scalar(z_type):
            schema_generate_scalar_filter(z_type, full_dict)
        else:
            schema_generate_type_filter(z_type, full_dict)
        assert full_dict["_Inputs"][filter_name] is not None 

def schema_generate_list_params(z_type, full_dict):
    name = z_type | F.Name | collect
    filter_name = f"{name}Filter"
    schema_generate_type_dispatch(z_type, full_dict)

    query_params = {
        "filter": {"type": filter_name},
        # We probably want to change these to be proper cursors.
        "first": {"type": "Int"},
        "offset": {"type": "Int"},
    }

    order_name = f"{name}Order"
    if order_name in full_dict["_Inputs"]:
        query_params["order"] = {"type": order_name}

    return query_params

def schema_generate_type_filter(z_type, full_dict):
    name = z_type | F.Name | collect
    fil_name = f"{name}Filter"
    list_fil_name = f"{name}FilterList"
    order_name = f"{name}Order"
    orderable_name = f"{name}Orderable"

    fields = {
        # TODO: I Think this is if the field is present?
        # "has: [{fil_name}]",
        "and": f"[{fil_name}]",
        "or": f"[{fil_name}]",
        "not": fil_name,
    }

    orderable_fields = []

    # Every type has an ID field, which is slightly custom - functions as an
    # automatically generated "in"
    fields["id"] = "[ID!]"

    for field in z_type | out_rels[RT.GQL_Field] | filter[op_is_searchable]:
        field_name = field | F.Name | collect
        field_type = target(field)
        field_type_name = field_type | F.Name | collect
        if is_a(field_type, AET.Bool):
            filter_name = "Boolean"
        else:
            filter_name = f"{field_type_name}Filter"
            schema_generate_type_dispatch(field_type, full_dict)

        if op_is_list(field):
            fields[field_name] = f"{filter_name}List"
        else:
            fields[field_name] = filter_name

    for field in z_type | out_rels[RT.GQL_Field] | filter[target | op_is_orderable]:
        orderable_fields += [field | F.Name | collect]

    full_dict["_Inputs"][fil_name] = fields

    if len(orderable_fields) > 0:
        full_dict["_Inputs"][order_name] = {
            "asc": orderable_name,
            "desc": orderable_name,
            "then": order_name,
        }
        full_dict["_Inputs"][list_fil_name] = {
            "any": fil_name,
            "all": fil_name,
            "size": "IntFilter",
        }
        orderable_dict = {x: x for x in orderable_fields}
        full_dict["_Enums"][orderable_name] = orderable_dict
        

def schema_generate_scalar_filter(z_node, full_dict):
    type_name = z_node | F.Name | collect
    fil_name = f"{type_name}Filter"
    list_fil_name = f"{type_name}FilterList"

    if is_a(z_node, AET.Bool):
        # Shoudln't need to do anything here, as it is true/false.
        return

    fil_dict = {}
    full_dict["_Inputs"][fil_name] = fil_dict

    fil_dict["eq"] = type_name
    fil_dict["in"] = f"[{type_name}]"

    if op_is_orderable(z_node):
        fil_dict["le"] = type_name
        fil_dict["lt"] = type_name
        fil_dict["ge"] = type_name
        fil_dict["gt"] = type_name
        fil_dict["between"] = f"{type_name}Range"

    if op_is_stringlike(z_node):
        fil_dict["contains"] = "String"

    full_dict["_Inputs"][list_fil_name] = {
        "any": fil_name,
        "all": fil_name,
        "size": "IntFilter",
    }
    
    if op_is_orderable(z_node):
        full_dict["_Inputs"][f"{type_name}Range"] = {
            "min": f"{type_name}!",
            "max": f"{type_name}!",
        }

####################################
# * Query resolvers
#----------------------------------

class ExternalError(Exception):
    pass

def resolve_get(_, info, *, type_node, **params):
    # Look for something that fits exactly what has been given in the params, assuming
    # that ariadne has done its work and validated the query.
    return find_existing_entity_by_id(info, type_node, params["id"])
@func
def resolve_get2(obj, type_node, graphql_info, query_args):
    return resolve_get(obj, graphql_info, type_node=type_node, **query_args)

def resolve_query(_, info, *, type_node, **params):
    ents = obtain_initial_list(type_node, params.get("filter", None), info)

    ents = handle_list_params(ents, type_node, params, info)

    return ents | collect

@func
def resolve_query2(obj, type_node, graphql_info, query_args):
    return resolve_query(obj, graphql_info, type_node=type_node, **query_args)

def resolve_aggregate(_, info, *, type_node, **params):
    # We can potentially defer the aggregation till later, by returning a kind
    # of lazy object here. However, for simplicity in the beginning, I will
    # aggregate everything, even if the query is only for a single field.

    ents = resolve_query(_, info, type_node=type_node, filter=params.get("filter", None))

    out = {"count": len(ents)}
    for z_field in type_node > L[RT.GQL_Field] | filter[op_is_aggregable]:
        vals = ents | map[lambda z: resolve_field(z, info, z_field=z_field)] | filter[Not[equals[None]]] | collect

        field_name = z_field | F.Name | collect

        if op_is_orderable(z_field | target):
            if len(vals) == 0:
                val_min = val_max = None
            else:
                val_min = min(vals)
                val_max = max(vals)

            # if isinstance(val_min, Time):
            #     val_min = str(val_min)
            #     val_max = str(val_max)

            out[f"{field_name}Min"] = val_min
            out[f"{field_name}Max"] = val_max

        if op_is_summable(z_field | target):
            if len(vals) == 0:
                val_sum = None
                val_avg = None
            else:
                val_sum = sum(vals)
                val_avg = val_sum / len(vals)

            out[f"{field_name}Sum"] = val_sum
            out[f"{field_name}Avg"] = val_avg

    return out
@func
def resolve_aggregate2(obj, type_node, graphql_info, query_args):
    return resolve_aggregate(obj, graphql_info, type_node=type_node, **query_args)

def resolve_field(z, info, *, z_field, **params):
    is_list = z_field | op_is_list | collect
    is_required = z_field | op_is_required | collect

    opts = internal_resolve_field(z, info, z_field)

    if is_list:
        opts = handle_list_params(opts, target(z_field), params, info)

    opts = collect(opts)

    if is_list:
        return opts
    else:
        if len(opts) == 0:
            if is_required:
                log.error("A single field has no option and is required!", z=z, z_field=z_field, z_name=fvalue(source(z_field), RT.Name, None), field_name=fvalue(z_field, RT.Name, None))
            return None
        if len(opts) >= 2:
            log.error("A single field has multiple options!", z=z, z_field=z_field, type_name=fvalue(source(z_field), RT.Name, None), field_name=fvalue(z_field, RT.Name, None))
            return None
        return single(opts)
@func
def resolve_field2(obj, z_field, graphql_info, query_args):
    return resolve_field(obj, graphql_info, z_field=z_field, **query_args)

def resolve_id(z, info):
    return str(origin_uid(z))

@func
def resolve_id2(obj, graphql_info):
    return resolve_id(obj, graphql_info)


##############################
# * Mutations
#----------------------------
from threading import Lock
mutation_lock = Lock()

def resolve_add(_, info, *, type_node, **params):
    # TODO: @unique checks should probably be done post change as multiple adds
    # could try changing the same thing, including nested types.
    try:
        with mutation_lock:
            name_gen = NameGen()
            actions = []
            post_checks = []
            new_obj_names = []
            updated_objs = []

            # This is not optimal but simplest to understand for now.
            upsert = params.get("upsert", False)
            for item in params["input"]:
                # Check id fields to see if we already have this.
                if "id" in item:
                    if not upsert:
                        raise ExternalError("Can't update item with id without setting upsert")
                    obj = find_existing_entity_by_id(info, type_node, item["id"])
                    if obj is None:
                        raise Exception("Item doesn't exist")
                    set_d = {**item}
                    set_d.pop("id")
                    new_actions,new_post_checks = update_entity(obj, info, type_node, set_d, {}, name_gen)
                    actions += new_actions
                    post_checks += new_post_checks
                    updated_objs += [obj]
                else:
                    obj_name,more_actions,more_post_checks = add_new_entity(info, type_node, item, name_gen)
                    actions += more_actions
                    post_checks += more_post_checks
                    new_obj_names += [obj_name]

            r = commit_with_post_checks(actions, post_checks, info)

            ents = updated_objs + [r[name] for name in new_obj_names]
            # Note: we return the details after any updates
            ents = ents | map[now] | collect

            count = len(new_obj_names)

            return {"count": count, "ents": ents}
    except ExternalError:
        raise
    except Exception as exc:
        if info.context["debug_level"] >= 0:
            log.error("There was an error in resolve_add", exc_info=exc)
        raise Exception("Unexpected error")
@func
def resolve_add2(obj, type_node, graphql_info, query_args):
    return resolve_add(obj, graphql_info, type_node=type_node, **query_args)
            
def resolve_upfetch(_, info, *, type_node, **params):
    # TODO: @unique checks should probably be done post change as multiple adds
    # could try changing the same thing, including nested types.
    try:
        with mutation_lock:
            # Upfetch is a lot like add with upsert, except it uses the specially
            # indicated upfetch field to work on, rather than id.
            name_gen = NameGen()
            actions = []
            post_checks = []
            new_obj_names = []
            updated_objs = []

            upfetch_field = op_upfetch_field(type_node)
            field_name = upfetch_field | F.Name | collect

            for item in params["input"]:
                # Check the upfetch field to see if we already have this.
                if field_name not in item:
                    raise Exception("Should never get here, because the upfetch field should be marked as required")

                obj = find_existing_entity_by_field(info, type_node, upfetch_field, item[field_name])
                if obj is None:
                    obj_name,more_actions,more_post_checks = add_new_entity(info, type_node, item, name_gen)
                    actions += more_actions
                    post_checks += more_post_checks
                    new_obj_names += [obj_name]
                else:
                    new_actions,new_post_checks = update_entity(obj, info, type_node, item, {}, name_gen)
                    actions += new_actions
                    post_checks += new_post_checks
                    updated_objs += [obj]

            r = commit_with_post_checks(actions, post_checks, info)

            ents = updated_objs + [r[name] for name in new_obj_names]
            # Note: we return the details after any updates
            ents = ents | map[now] | collect
            count = len(new_obj_names)

            return {"count": count, "ents": ents}
    except ExternalError:
        raise
    except Exception as exc:
        if info.context["debug_level"] >= 0:
            log.error("There was an error in resolve_upfetch", exc_info=exc)
        raise Exception("Unexpected error")
@func
def resolve_upfetch2(obj, type_node, graphql_info, query_args):
    return resolve_upfetch(obj, graphql_info, type_node=type_node, **query_args)
            
        
def resolve_update(_, info, *, type_node, **params):
    # TODO: @unique checks should probably be done post change as multiple adds
    # could try changing the same thing, including nested types.
    try:
        with mutation_lock:
            if "input" not in params or "filter" not in params["input"]:
                raise Exception("Not allowed to update everything!")
            ents = resolve_query(_, info, type_node=type_node, filter=params["input"]["filter"])

            actions = []
            post_checks = []

            name_gen = NameGen()
            for ent in ents:
                new_actions,new_post_checks = update_entity(ent, info, type_node, params["input"].get("set", {}), params["input"].get("remove", {}), name_gen)
                actions += new_actions
                post_checks += new_post_checks

            commit_with_post_checks(actions, post_checks, info)

            count = len(ents)

            # Note: we return the details after the update
            ents = ents | map[now] | collect

            return {"count": count, "ents": ents}
    except ExternalError:
        raise
    except Exception as exc:
        if info.context["debug_level"] >= 0:
            log.error("There was an error in resolve_update", exc_info=exc)
        raise Exception("Unexpected error")
@func
def resolve_update2(obj, type_node, graphql_info, query_args):
    return resolve_update(obj, graphql_info, type_node=type_node, **query_args)

def resolve_delete(_, info, *, type_node, **params):
    try:
        with mutation_lock:
            # Do the same thing as a resolve_query but delete the entities instead.
            if "filter" not in params:
                raise ExtenalError("Not allowed to delete everything!")
            ents = resolve_query(_, info, type_node=type_node, **params)

            if not ents | map[pass_delete_auth[type_node][info]] | all | collect:
                raise ExtenalError("Auth check returned False")

            post_checks = []
            for ent in ents:
                post_checks += [("remove", ent, type_node)]
            actions = [terminate(ent) for ent in ents]
            r = commit_with_post_checks(actions, post_checks, info)

            count = len(ents)

            return {"count": count, "ents": ents}
    except ExternalError:
        raise
    except Exception as exc:
        if info.context["debug_level"] >= 0:
            log.error("There was an error in resolve_delete", exc_info=exc)
        raise Exception("Unexpected error")
@func
def resolve_delete2(obj, type_node, graphql_info, query_args):
    return resolve_delete(obj, graphql_info, type_node=type_node, **query_args)

def resolve_filter_response(obj, info, *, type_node, **params):
    ents = obj["ents"]

    ents = handle_list_params(ents, type_node, params, info)

    return ents | collect
@func
def resolve_filter_response2(obj, type_node, graphql_info, query_args):
    return resolve_filter_response(obj, graphql_info, type_node=type_node, **query_args)




##############################################
# * Internal query parts
#--------------------------------------------

def obtain_initial_list(type_node, filter_opts, info):
    gs = info.context["gs"]

    type_et = ET(type_node | Out[RT.GQL_Delegate] | collect)
    if filter_opts is not None and filter_opts.get("id", None) is not None:
        # ids were provided, so the initial list starts with them

        # Note: we make the decision here to not throw on a missing id, as that
        # could potentially be exploited somehow by smoeone to learn about what
        # items exist even if they don't have auth access. A missing id is hence
        # just "not passing" the filter.

        ids = filter_opts["id"]
        zs = []

        zs = (ids
              | map[lambda id: find_existing_entity_by_id(info, type_node, id)]
              | filter[Not[equals[None]]])

        if info.context["debug_level"] >= 3:
            log.debug("DEBUG 3: built initial list from ids", length_ids=len(ids), length_list=length(zs))

        return zs
    else:
        zs = gs | all[type_et] | filter[pass_query_auth[type_node][info]]
        if info.context["debug_level"] >= 3:
            log.debug("DEBUG 3: built initial list from type and auth", length_list=length(zs))
        return zs

    

def handle_list_params(opts, z_node, params, info):
    opts = maybe_filter_result(opts, z_node, info, params.get("filter", None))
    if info.context["debug_level"] >= 3:
        log.debug("DEBUG 3: after filtering", length_list=length(opts))
    opts = maybe_sort_result(opts, z_node, info, params.get("order", None))
    opts = maybe_paginate_result(opts, params.get("first", None), params.get("offset", None))
    return opts

@func
def field_resolver_by_name(z, type_node, info, name):
    # Note name goes last because it is curried last... this is weird
    if name == "id":
        return resolve_id(z, info=info)
    sub_field = get_field_rel_by_name(type_node, name)
    return resolve_field(z, info=info, z_field=sub_field)


# ** Filtering

def maybe_filter_result(opts, z_node, info, fil=None):
    if fil is None:
        return opts

    temp_fil = build_filter_zefop(fil, z_node, info)
    if info.context["debug_level"] >= 4:
        log.debug("DEBUG 4: filter is", fil=temp_fil)
        temp_fil = (apply[identity, temp_fil]
                    | tap[match[
                        (Is[second], lambda x: log.debug("DEBUG 4: filter passed", item=first(x))),
                        (Any, lambda x: log.debug("DEBUG 4: filter failed", item=first(x)))
                        ]]
                    | second)
    return opts | filter[temp_fil]

def build_filter_zefop(fil, z_node, info):
    field_resolver = field_resolver_by_name[z_node][info]
    # top level is ands
    top = And
    for key,sub in fil.items():
        if key == "and":
            this = And
            for part in sub:
                this = this[build_filter_zefop(part, z_node, info)]
        elif key == "or":
            this = Or
            for part in sub:
                this = this[build_filter_zefop(part, z_node, info)]
        elif key == "not":
            this = Not[build_filter_zefop(sub, z_node, info)]
        elif key == "id":
            # This is handled specially - functions like an "in".
            val = field_resolver["id"]
            this = val | contained_in[sub]
        else:
            # This must be a field
            z_field = get_field_rel_by_name(z_node, key)
            z_field_node = target(z_field)

            val = field_resolver[key]
            if op_is_list(z_field):
                list_top = And
                for list_key,list_sub in sub.items():
                    if list_key == "any":
                        sub_fil = build_filter_zefop(list_sub, z_field_node, info)
                        list_top = list_top[val | map[sub_fil] | any]
                    elif list_key == "all":
                        sub_fil = build_filter_zefop(list_sub, z_field_node, info)
                        list_top = list_top[val | map[sub_fil] | all]
                    elif list_key == "size":
                        temp = val | length | scalar_comparison_op(list_sub)
                        list_top = list_top[temp]
                    else:
                        raise Exception(f"Unknown list filter keyword: {list_key}")
                this = list_top
            elif op_is_scalar(z_field_node):
                if isinstance(sub, bool):
                    this = val | equals[sub]
                else:
                    this = val | scalar_comparison_op(sub)
            else:
                sub_fil = build_filter_zefop(sub, z_field_node, info)
                this = val | And[Not[equals[None]]][sub_fil]


        top = top[this]

    return top

def scalar_comparison_op(sub):
    this = And[Not[equals[None]]]
    for scalar_key,scalar_sub in sub.items():
        if scalar_key == "eq":
            this = this[equals[scalar_sub]]
        elif scalar_key == "in":
            this = this[contained_in[scalar_sub]]
        elif scalar_key == "contains":
            this = this[contains[scalar_sub]]
        elif scalar_key == "le":
            this = this[less_than_or_equal[scalar_sub]]
        elif scalar_key == "lt":
            this = this[less_than[scalar_sub]]
        elif scalar_key == "ge":
            this = this[greater_than_or_equal[scalar_sub]]
        elif scalar_key == "gt":
            this = this[greater_than[scalar_sub]]
        elif scalar_key == "between":
            this = this[greater_than_or_equal[scalar_sub["min"]]]
            this = this[less_than_or_equal[scalar_sub["max"]]]
        else:
            raise Exception(f"Unknown comparison operator: {key}")
    return this
                
def get_field_rel_by_name(z_type, name):
    return (z_type | out_rels[RT.GQL_Field]
            | filter[Z | F.Name | equals[name]]
            | first
            | collect)

# ** Sorting

def maybe_sort_result(opts, z_node, info, sort_decl=None):
    if sort_decl is None:
        return opts

    field_resolver = field_resolver_by_name[z_node][info]

    # First, get the list of things to sort by, so that we can reverse it
    sort_list = []
    cur = sort_decl
    while cur is not None:
        if "asc" in cur:
            assert not "desc" in cur
            val = field_resolver[cur["asc"]]
            sort_list += [sort[val]]
        elif "desc" in cur:
            val = field_resolver[cur["desc"]]
            sort_list += [sort[val][True]]

        cur = cur.get("then", None)

    sort_list.reverse()

    for action in sort_list:
        opts = opts | action

    return opts

# ** Pagination

def maybe_paginate_result(opts, first=None, offset=None):
    if offset is not None:
        opts = opts | skip[offset]
    if first is not None:
        opts = opts | take[first]
    return opts


# ** Resolution

@func
def internal_resolve_field(z, info, z_field, auth_required=True):
    # This returns a LazyValue so we can deal with whatever comes out without
    # instantiating the whole list.

    # Also *only* returns that list. The calling function needs to apply the single/list logic
    
    is_incoming = z_field | op_is_incoming | collect
    # This is a delegate
    if z_field | has_out[RT.GQL_Resolve_With] | collect:
        relation = z_field | Out[RT.GQL_Resolve_With] | collect
        is_triple = source(relation) != relation

        rt = RT(relation)

        if is_triple:
            if is_incoming:
                assert rae_type(z) == rae_type(target(relation)), f"The RAET of the object {z} is not the same as that of the delegate relation {target(relation)}"
            else:
                assert rae_type(z) == rae_type(source(relation)), f"The RAET of the object {z} is not the same as that of the delegate relation {source(relation)}"

        if is_incoming:
            opts = z | Ins[rt]
            if is_triple:
                opts = opts | filter[is_a[rae_type(source(relation))]]
        else:
            opts = z | Outs[rt]
            if is_triple:
                opts = opts | filter[is_a[rae_type(target(relation))]]
    elif z_field | has_out[RT.GQL_FunctionResolver] | collect:
        opts = func[z_field | Out[RT.GQL_FunctionResolver] | collect](z, info)
        # This is to mimic the behaviour that people probably expect from a
        # non-list resolver.
        if type(opts) not in [list,tuple]:
            if opts is None:
                opts = []
            else:
                opts = [opts]
    else:
        raise Exception(f"Don't know how to resolve this field: {z_field}")

    if auth_required:
        opts = opts | filter[pass_query_auth[target(z_field)][info]]

    # We must convert final objects from AEs to python types.
    # if z_field | target | is_core_scalar | collect:
    if z_field | target | op_is_scalar | collect:
        # With a dynamic resolver, the items could also be values, so only apply value if they are ZefRefs
        opts = opts | map[match[
            (ZefRef, value),
            (Any, identity)
        ]]

    return opts


# ** Handling adding

def NameGen():
    n = 0
    while True:
        n += 1
        yield n

def find_existing_entity_by_id(info, type_node, id):
    if id is None:
        return None
    the_uid = uid(id)
    if the_uid is None:
        raise Exception("An id of {id} cannot be converted to a uid.")
    
    gs = info.context["gs"]

    et = ET(type_node | Out[RT.GQL_Delegate] | collect)
    ent = gs[uid(id)] | collect
    if not is_a(ent, et):
        return None
    if not ent | pass_query_auth[type_node][info] | collect:
        return None

    return ent

def find_existing_entity_by_field(info, type_node, z_field, val):
    if val is None:
        raise Exception("Can't find an entity by a None field value")

    gs = info.context["gs"]
    et = ET(type_node | Out[RT.GQL_Delegate] | collect)

    # Note: we filter out with query auth even though this may mean we return
    # None instead of the actual entity. The follow-up logic with an upfetch
    # might seem wrong, as it would try to create the same item again. However,
    # because the field is also @unique then this will fail. Better to keep the
    # failure point at the same location so we can determine what kind of error
    # to return and whether this will leak sensitive information.
    ent = (gs | all[et] | filter[pass_query_auth[type_node][info]]
           | filter[internal_resolve_field[info][z_field] | optional | equals[val]]
           | optional
           | collect)

    return ent

def add_new_entity(info, type_node, params, name_gen):

    actions = []
    post_checks = []

    this = str(next(name_gen))
    type_name = type_node | F.Name | collect

    post_checks += [("add", this, type_node)]

    et = ET(type_node | Out[RT.GQL_Delegate] | collect)
    actions += [et[this]]

    # This should probably be cached
    field_mapping = {}
    for z_field in type_node | out_rels[RT.GQL_Field]:
        field_name = z_field | F.Name | collect
        field_mapping[field_name] = z_field
        # Convenient spot to validate that required fields have been specified.
        if op_is_required(z_field):
            if field_name not in params:
                raise ExternalError(f"Required field '{field_name}' not given when creating new entity of type '{type_name}'")


    for key,val in params.items():
        # TODO: Validate that any unique field is not duplicated
        z_field = field_mapping[key]
        field_name = z_field | F.Name | collect
        rt = RT(z_field | Out[RT.GQL_Resolve_With] | collect)
        if z_field | op_is_list | collect:
            if not isinstance(val, list):
                raise Exception(f"Value should have been list but was {type(val)}")

        if z_field | target | op_is_scalar | collect:
            if op_is_unique(z_field):
                others = info.context["gs"] | all[et] | filter[fvalue[rt][None] | equals[val]] | func[set] | collect
                if len(others) > 0:
                    if info.context["debug_level"] >= 0:
                        log.error("Trying to add a new entity with unique field that conflicts with others", et=et, field=field_name, others=others)
                    raise ExternalError(f"Unique field '{field_name}' conflicts with existing items.")
            if z_field | op_is_list | collect:
                l = val
            else:
                l = [val]

            for item in l:
                # TODO: Maybe proper ordering here too? Or should this
                # always be considered to be a set, ordered only when the
                # client requests it?
                if z_field | op_is_incoming | collect:
                    actions += [item, rt, (Z[this])]
                else:
                    actions += [(Z[this], rt, item)]
        else:
            if z_field | op_is_list | collect:
                l = val
            else:
                l = [val]

            for item in l:
                obj,obj_actions,obj_post_checks = find_or_add_entity(item, info, target(z_field), name_gen)
                actions += obj_actions
                post_checks += obj_post_checks
                if z_field | op_is_incoming | collect:
                    actions += [(obj, rt, Z[this])]
                else:
                    actions += [(Z[this], rt, obj)]

    return this, actions, post_checks

def find_or_add_entity(val, info, target_node, name_gen):
    if isinstance(val, dict) and val.get("id", None) is not None:
        # There should be no other fields given for this entity, otherwise the meaning is unclear.
        if set(val.keys()) != {"id"}:
            raise ExternalError("An entity should be designated with either an 'id' or a set of new fields to create a new entity.")
        obj = find_existing_entity_by_id(info, target_node, val["id"])
        if obj is None:
            raise ExternalError(f"Unable to find entity of kind '{target_node | F.Name | collect}' with id '{val['id']}'.")
        return obj,[],[]
    else:
        obj_name,actions,post_checks = add_new_entity(info, target_node, val, name_gen)
        return Z[obj_name], actions, post_checks
    

def update_entity(z, info, type_node, set_d, remove_d, name_gen):
    type_name = type_node | F.Name | collect

    # Refuse to set and remove the same thing. Just too confusing to deal with
    assert(len(set(set_d.keys()).intersection(remove_d.keys())) == 0), "Can't have the same set/remove keys"
    if len(set_d) + len(remove_d) == 0:
        raise ExternalError("No set or remove in update_entity!")

    # This is the pre-update auth only
    # TODO: Post-update auth
    if not pass_pre_update_auth(z, type_node, info):
        raise ExternalError("Not allowed to update")

    actions = []
    post_checks = []
    post_checks += [("update", z, type_node)]

    # This should probably be cached
    field_mapping = {}
    for z_field in type_node | out_rels[RT.GQL_Field]:
        field_mapping[z_field | F.Name | collect] = z_field

    for key,val in set_d.items():
        # TODO: Validate that any unique field is not duplicated
        z_field = field_mapping[key]
        field_name = z_field | F.Name | collect
        # TODO: This should be able to distinguish based on the triple, not just the RT
        rt = RT(z_field | Out[RT.GQL_Resolve_With] | collect)

        if op_is_list(z_field):
            raise Exception(f"Updating list things is a todo (for z_field={z_field})")
        else:
            if op_is_unique(z_field):
                # This used to be checked directly, here but just in case
                # there's a mutation that renames two things at the same time,
                # we change this up a bit.
                post_checks += [("unique", z_field, type_node)]

            if z_field | target | op_is_scalar | collect:
                actions += [z | set_field[rt][val][op_is_incoming(z_field)]]
            else:
                found_z,new_actions,new_post_checks = find_or_add_entity(val, info, target(z_field), name_gen)
                actions += new_actions
                post_checks += new_post_checks

                actions += [z | set_field[rt][found_z][op_is_incoming(z_field)]]
    
    for key,val in remove_d.items():
        if val is not None:
            raise ExternalError("Remove vals need to be nil")
        z_field = field_mapping[key]
        field_name = z_field | F.Name | collect
        if op_is_required(z_field):
            raise ExternalError(f"Not allowed to remove required field '{field_name}' on type '{type_name}'")
        # TODO: This should be able to distinguish based on the triple, not just the RT
        rt = RT(z_field | Out[RT.GQL_Resolve_With] | collect)
        if op_is_incoming(z_field):
            rels = z | in_rels[rt]
        else:
            rels = z | out_rels[rt]
        actions += [terminate(rel) for rel in rels]

        # TODO: Add a post-check here that this would not remove a relation on a
        # type that has a required relation. Note that a new relation could be
        # created in the same transaction, so it needs to be post, not directly
        # here.

        # Also delete scalars
        if z_field | target | op_is_scalar | collect:
            actions += [terminate(target(rel)) for rel in rels]

    return actions, post_checks


##############################
# * Auth things
#----------------------------

def pass_auth_generic(z, schema_node, info, rt_list):
    to_call = None
    for rt in rt_list:
        to_call = schema_node | Outs[rt] | optional | collect
        if to_call is not None:
            break
    else:
        return True

    # TODO: Later on these will be zef functions so easier to call
    return temporary__call_string_as_func(
        to_call | value | collect,
        z=z,
        info=info,
        type_node=schema_node
    )

@func
def pass_query_auth(z, schema_node, info):
    return pass_auth_generic(z, schema_node, info, [RT.AllowQuery])

@func
def pass_add_auth(z, schema_node, info):
    return pass_auth_generic(z, schema_node, info, [RT.AllowAdd, RT.AllowQuery])

@func
def pass_pre_update_auth(z, schema_node, info):
    return pass_auth_generic(z, schema_node, info, [RT.AllowUpdate, RT.AllowQuery])

@func
def pass_post_update_auth(z, schema_node, info):
    return pass_auth_generic(z, schema_node, info, [RT.AllowUpdatePost, RT.AllowUpdate, RT.AllowQuery])

@func
def pass_delete_auth(z, schema_node, info):
    return pass_auth_generic(z, schema_node, info, [RT.AllowDelete, RT.AllowUpdate, RT.AllowQuery])

def temporary__call_string_as_func(s, **kwds):
    from ... import core, ops
    try:
        out = eval(
            s, {
                **{name: getattr(core,name) for name in dir(core) if not name.startswith("_")},
                **{name: getattr(ops,name) for name in dir(ops) if not name.startswith("_")},
                "auth_field": P(auth_helper_auth_field, **kwds),
            },
            kwds
        )
    except Exception as exc:
        log.error("Problem calling auth expression", expr=s, exc_info=exc, kwds=kwds)
        return False

    if isinstance(out, LazyValue):
        out = collect(out)
    elif isinstance(out, ZefOp):
        out = out(kwds.get("z", None))

    return out

def auth_helper_auth_field(field_name, auth, *, z, type_node, info):
    # A helper function for graphql schema, that requests an auth check of the
    # given kind on one of its fields.

    # Only makes sense for fields that are required, such as a user field.
    try:
        z_field = get_field_rel_by_name(type_node, field_name)
        z_field_node = target(z_field)
        val = field_resolver_by_name(z, type_node, info, field_name)
    except:
        # Going to assume this is because traversal failed auth along the way somewhere.
        if info.context["debug_level"] >= 0:
            log.error("auth_field helper got an exception, assuming failure of auth")
        return False

    if val is None:
        # This is something we can't query on, so therefore the query has failed.
        return False

    if auth == "query":
        func = pass_query_auth
    elif auth == "add":
        func = pass_add_auth
    elif auth == "update":
        func = pass_pre_update_auth
    elif auth == "updatePost":
        func = pass_post_update_auth
    elif auth == "delete":
        func = pass_delete_auth
    else:
        raise Exception(f"Don't understand auth type '{auth}' in auth_helper_auth_field")

    return func(val, z_field_node, info)



def commit_with_post_checks(actions, post_checks, info):
    g = Graph(info.context["gs"])
    with Transaction(g):
        try:
            r = actions | transact[g] | run
            # Test all post checks
            for (kind,obj,type_node) in post_checks:
                type_name = type_node | fvalue[RT.Name][''] | collect

                if kind in ["unique"]:
                    pass
                else:
                    if type(obj) == str:
                        obj = r[obj]
                    assert type(obj) == ZefRef
                    obj = obj | in_frame[g | now | collect][allow_tombstone] | collect

                if kind == "add":
                    for z_func in type_node | Outs[RT.OnCreate]:
                        try:
                            func[z_func](obj)
                        except:
                            raise ExternalError(f"OnCreate hook for type_node of '{type_name}' threw an exception")
                    if not pass_add_auth(obj, type_node, info):
                        raise ExternalError(f"Add auth check for type_node of '{type_name}' returned False")
                elif kind == "update":
                    for z_func in type_node | Outs[RT.OnUpdate]:
                        try:
                            func[z_func](obj)
                        except:
                            raise ExternalError(f"OnUpdate hook for type_node of '{type_name}' threw an exception")
                    if not pass_post_update_auth(obj, type_node, info):
                        raise ExternalError(f"Post-update auth check for type_node of '{type_name}' returned False")
                elif kind == "remove":
                    for z_func in type_node | Outs[RT.OnRemove]:
                        try:
                            func[z_func](obj)
                        except:
                            raise ExternalError(f"OnRemove hook for type_node of '{type_name}' threw an exception")
                elif kind == "unique":
                    # In this case, obj is the field which must be unique
                    z_field = obj
                    assert op_is_unique(z_field)
                    # Get all values - note this is not filtered by the user's
                    # viewpoint, so we need to be a little careful.
                    ents = g | now | all[ET(type_node | Out[RT.GQL_Delegate] | collect)]
                    vals = ents | map[internal_resolve_field[info][z_field][False]] | concat | collect

                    dis = distinct(vals)
                    if len(dis) != len(vals):
                        if info.context["debug_level"] >= 0:
                            log.error("Non-unique values", vals=set(vals) - set(dis))
                        raise ExternalError(f"Non-unique values found for field '{z_field | F.Name | collect}' of type_node '{type_name}'")


        except Exception as exc:
            if info.context["debug_level"] >= 0:
                log.error("Aborting transaction",
                          exc_info=exc)
            from ...pyzef.internals import AbortTransaction
            AbortTransaction(g)
            if type(exc) == ExternalError:
                raise exc
            else:
                raise ExternalError("Unexpected error in auth check/hooks execution")

    return r



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


#                           _   _                  __     __        _               _____                                             
#                          | | | | ___   ___  _ __ \ \   / /  __ _ | | _   _   ___ |_   _| _   _  _ __    ___                         
#   _____  _____  _____    | | | |/ __| / _ \| '__| \ \ / /  / _` || || | | | / _ \  | |  | | | || '_ \  / _ \    _____  _____  _____ 
#  |_____||_____||_____|   | |_| |\__ \|  __/| |     \ V /  | (_| || || |_| ||  __/  | |  | |_| || |_) ||  __/   |_____||_____||_____|
#                           \___/ |___/ \___||_|      \_/    \__,_||_| \__,_| \___|  |_|   \__, || .__/  \___|                        
#                                                                                          |___/ |_|                                  

from .VT import String, ValueType, Dict, Any
from ._ops import is_a


_user_value_type_registry = {}    # append to this: keep track of all types known to this runtime

class UserValueType_:
    def __init__(self, name: String, representation_type, constraints):
        from random import randint
        self.name = name
        self.representation_type = representation_type
        self.constraints = constraints
        self.user_type_id = str(randint(0, 100000000000))     # TODO: use a Ref: the type is originally created on the respective process graph

        # add to registry
        _user_value_type_registry[self.user_type_id] = self

    def __repr__(self):
        return f"UserValueType(name={self.name}, representation_type={self.representation_type}, constraints={self.constraints})"

    def __call__(self, *args, **kwargs):
        """
        equivalent to the constructor for this custom type
        """
        
        if bool(kwargs):
            # if keyword args, then the representation_type MUST be a dict and no positional args are allowed
            assert args == ()
            assert self.representation_type in {Dict, dict}
            val = kwargs

        else:
            # if positional args, then the representation_type MUST be a tuple and no keyword args are allowed
            assert bool(kwargs)==False    # no keyword args
            if len(args) == 1:         # if there is only one positional arg, then it is the value itself
                val = args[0]
            elif len(args) == 0 and self.representation_type in {Dict, dict}:
                # The constructor is valid without args only in the case of a dictionary
                val = {}
            else:
                raise ValueError("Error initializing UserValueType")

        try:
            cast_val = self.representation_type(val)
        except:
            raise Exception(f"UserValueType(name={self.name}) cannot cast '{val}' to {self.representation_type}")

        # check whether the constraints are satisfied
        if not is_a(cast_val, self.constraints):
            raise Exception(f"UserValueType(name={self.name}) constraint does not match")
        
        return UserValueInstance_(self.user_type_id, cast_val)
        
    def __eq__(self, other):
        if type(other) is not type(self):
            return False
        if other.user_type_id != self.user_type_id:
            return False
        return True


# use a constructor function distinct from the type
def UserValueType(name: String, representation_type: ValueType, constraints: ValueType=Any):
    return UserValueType_(name, representation_type, constraints)



class UserValueInstance_:
    def __init__(self, user_type_id, value):
        # only store the id here, not the name?
        self.user_type_id = user_type_id
        self.value = value
        
    def __repr__(self):
        # look up the name in the _user_value_type_registry from the id
        type_name = _user_value_type_registry[self.user_type_id].name
        return f"{type_name}({repr(self.value)})"

    def __eq__(self, other):
        if type(other) is not type(self):
            return False
        if other.user_type_id != self.user_type_id:
            return False
        if other.value != self.value:
            return False
        return True

    def __getattr__(self, other):
        if not isinstance(self.value, dict):
            raise AttributeError("UserValueInstance 'dot' access only works on dicts")
        return self.value[other]

    # required to enable dict(my_user_value_type_instance)
    # fingers crossed that nobody has a field 'keys' 
    # and tries to do my_user_value_type_instance.keys
    def keys(self):
        if not isinstance(self.value, dict):
            raise AttributeError("UserValueInstance 'keys' only works on dicts")
        return self.value.keys()

    def __getitem__(self, key):
        if not isinstance(self.value, dict):
            raise AttributeError("UserValueInstance 'getitem' only works on dicts")
        return self.value[key]



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from zef import *
from zef.ops import *
from .ariadne_utils import gqlify, de_gqlify
from ..core.logger import log

# TODO attach Delegates


# General Utils #
def find_field_with_name(gql_type, name):
    for f in gql_type > L[RT(gqlify("field"))] | collect:
        if de_gqlify(f >> RT.Name | value | collect, False) == name:
            return f


def connect_direct_resolvers_to_delegates(g, gql_type, data_type, field_to_rt):
    print("#TODO This is deprecated use connect_delegate_resolvers instead")
    for field_name, rt_name in field_to_rt.items():
        field_zr = find_field_with_name(gql_type, field_name) | to_ezefref | collect
        data_delegate = ET(data_type) | delegate_of[g] > RT(rt_name) | collect
        create_resolver_to_field(g, field_zr, data_delegate)

# This is deprecated use connect_delegate_resolvers instead
def connect_direct_resolvers_to_specific_delegate(g, gql_type, data_type, tups_of_specific_relations):
    print("#TODO This is deprecated use connect_delegate_resolvers instead")
    for tup in tups_of_specific_relations:
        field_name, rt_name, target_type = tup
        field_zr = find_field_with_name(gql_type, field_name) | to_ezefref | collect
        data_delegate = (ET(data_type) | delegate_of[g] > L[RT(rt_name)] | collect) | filter[lambda x: x | target | AET | collect == target_type] | only | collect 
        create_resolver_to_field(g, field_zr, data_delegate)


def connect_zef_function_resolvers(g, gql_type, field_resolving_dict):
    for field_name, handler_func in field_resolving_dict.items():
        gql_rt = find_field_with_name(gql_type, field_name)
        create_zef_function_resolver(g, gql_rt, handler_func)

def connect_delegate_resolvers(g, gql_type, field_resolving_dict):
    for field_name, delegate_details in field_resolving_dict.items():
        delegate_triple, is_out = delegate_details["triple"], delegate_details.get("is_out", True)
        field_zr = find_field_with_name(gql_type, field_name) | to_ezefref | collect
        data_delegate = [delegate_of(delegate_triple)] | transact[g] | run | get[delegate_of(delegate_triple)] | collect
        create_resolver_to_field(g, field_zr, data_delegate, is_out)

# Resolve With Connectors#
def create_zefscript_resolver(g, gql_rt, script):
    if length(now(gql_rt) >> L[RT(gqlify("resolve_with_script"))] | collect) == 0:
        aet_str = instantiate(AET.String, g)
        (aet_str <= script) | g | run
        zef_scr = instantiate(ET.ZEF_Script, g)
        instantiate(zef_scr, RT.ZEF_Python, aet_str, g)
        instantiate(gql_rt, RT(gqlify("resolve_with_script")), zef_scr, g)
    else:
        log.warn(f"The bindings for resolve_with_script on {gql_rt | uid} has been updated")
        (((gql_rt >> RT(gqlify("resolve_with_script")) | collect) >> RT.ZEF_Python) <= script) | g | run

def create_body_resolver(g, gql_rt, script):
    if length(now(gql_rt) >> L[RT(gqlify("resolve_with_body"))]) == 0:
        aet_str = instantiate(AET.String, g)
        (aet_str <= script) | g | run
        instantiate(gql_rt, RT(gqlify("resolve_with_body")), aet_str, g)
    else:
        log.warn(f"The bindings for resolve_with_body on {gql_rt | uid} has been updated")
        ((gql_rt >> RT(gqlify("resolve_with_body")) | collect ) <= script) | g | run


def create_intermediate_resolver_to_field(g, gql_rt, del_rt):
    if length(now(gql_rt) >> L[RT(gqlify("resolve_with_intermediate"))]) == 0:
        gql_rt, del_rt = gql_rt | to_ezefref | collect, del_rt | to_ezefref | collect
        instantiate(gql_rt, RT(gqlify("resolve_with_intermediate")), del_rt, g)


def create_resolver_to_field(g, gql_rt, del_rt, is_out = False):
    if length(now(gql_rt) >> L[RT(gqlify("resolve_with"))]) == 0:
        gql_rt, del_rt = gql_rt | to_ezefref | collect, del_rt | to_ezefref | collect
        rt = instantiate(gql_rt, RT(gqlify("resolve_with")), del_rt, g)
        aet = instantiate(AET.Bool, g)
        (aet <= is_out) | g | run
        rt = instantiate(rt, RT.IsOut, aet, g)


def create_zef_function_resolver(g, gql_rt, z_func):
    if length(now(gql_rt) >> L[RT(gqlify("resolve_with_zef_function"))]) == 0:
        (gql_rt, RT(gqlify("resolve_with_zef_function")), z_func) | g | run

# Relay #
def apply_cursor_to_edges(edges, before=None, after=None):
    def check_edge_location(edges, cursor, return_cursor):
        for idx, elm in enumerate(edges):
            if elm | uid == cursor:
                return idx
        return return_cursor

    prior_to_after = False
    if after:
        after_Edge = check_edge_location(edges, after, 0)
        if after_Edge != 0:
            edges = edges[after_Edge + 1:]
            prior_to_after = True

    following_before = False
    if before:
        before_edge = check_edge_location(edges, before, length(edges))
        if before_edge != length(edges):
            edges = edges[:before_edge]
            following_before = True

    return edges, prior_to_after, following_before


def edges_to_return(edges, first=None, last=None):
    if first and last:
        raise Exception("Both first and last are passed")

    if first:
        if first < 0:
            raise Exception("First is less than 0")

        if length(edges) > first:
            edges = edges[:first]

    if last:
        if last < 0:
            raise Exception("last is less than 0")

        if length(edges) > last:
            edges = edges[length(edges) - last:]

    return edges


def has_previous(edges, prior_to_after, after=None, last=None):
    if last:
        if length(edges) > last:
            return True
        return False

    elif after:
        # If the server can efficiently determine that elements exist prior to after, return true.
        return prior_to_after
    else:
        return False


def has_next(edges, following_before, before=None, first=None):
    if first:
        if length(edges) > first:
            return True
        return False

    elif before:
        # If the server can efficiently determine that elements exist following before, return true.
        return following_before
    else:
        return False


def pagination_algorithm(edges, before=None, after=None, first=None, last=None):
    if not edges:
        return {"pageInfo": {
            "hasPreviousPage": False,
            "hasNextPage": False,
            "startCursor": "",
            "endCursor": ""
        },
            "edges": []}
    edges = [e for e in edges]
    edges, prior_to_after, following_before = apply_cursor_to_edges(edges, before, after)
    return_edges = edges_to_return(edges, first, last)

    return {"pageInfo": {
        "hasPreviousPage": has_previous(edges, prior_to_after, after, last),
        "hasNextPage": has_next(edges, following_before, before, first),
        "startCursor": return_edges[0] | uid,
        "endCursor": return_edges[length(return_edges) - 1] | uid
    },
        "edges": return_edges}

